import copy
import anndata
import numpy as np
import pandas as pd
import anndata as ad
import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib.colors import LinearSegmentedColormap, ListedColormap
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from os import PathLike
from types import MappingProxyType
from typing import Optional, Union, Any, Literal, Tuple, Mapping, Sequence
from ._heatmap import _heatmap_with_dendrogram_and_bar
from ..utils import _get_info_from_sample, _scale
from ._color import _get_palette

__all__ = [
    'show_celltype_sample_heatmap',
    'show_niche_environment',
    'show_celltype_niche_heatmap',
    'show_niche_sample_heatmap'
]


def show_celltype_sample_heatmap(
        adata: anndata.AnnData,
        norm_axis: Literal['celltype', 'sample', None] = None,
        norm_method: Literal['normalization', 'z_score', 'proportion'] = 'normalization',
        celltype_bar: bool = True,
        sample_bar: bool = True,
        title: Optional[str] = None,
        figsize: Optional[Tuple[float, float]] = None,
        dpi: int = 100,
        cmap: LinearSegmentedColormap = None,
        show: bool = True,
        save: Union[str, PathLike] = None,
        celltype_bar_kwargs: dict = {},
        sample_bar_kwargs: dict = {},
        **kwargs
) -> Optional[Tuple[Figure, dict]]:
    """
    Heatmap show the cell composition of each sample. Bar plot can be added to show the overall cell types distribution
    and the total number of cells in each sample, respectively.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    norm_axis : Literal['celltype', 'sample']
        The direction of standardization.
        'celltype': Standardization was performed for each cell type.
        'sample': Standardization was performed for each sample.
        None: No standardization.
    celltype_bar
        Whether to show a bar plot of the cell number of each cell type.
    sample_bar
        Whether to show a bar plot of the cell number of each sample.
    celltype_bar_kwargs : Any
        Other parameters in sns.countplot() of cell type's bar plot.
    sample_bar_kwargs : Any
        Other parameters in sns.countplot() of sample's bar plot.
    norm_method : Literal['normalization', 'z_score', 'proportion']
        Methods of standardization
        'normalization': x^ = (x - min(x)) / (max(x) - min(x))
        'z_score': x^ = (x - mean(x)) / std(x)
        'proportion': x^ = x / sum(x)
    figsize : Tuple[float, float], optional
        (Width, height) of the figure.
    dpi : float, optional
        The resolution of the figure.
    title : str, optional
        The title of shown figure.
    show : bool
        Show the plot, do not return axis.
    save : Union[str, PathLike], optional
        The path where the image is stored.
    cmap
        Color map to use for continous variables.
    kwargs : Any
        Other params of sns.heatmap()

    Returns
    -------
        If show==False, return Tuple[Figure, list[Axes]]

    """
    if celltype_bar_kwargs is None:
        celltype_bar_kwargs = {}
    if sample_bar_kwargs is None:
        sample_bar_kwargs = {}

    if title is None:
        title = 'Celltype-Sample'

    norm_axis = 0 if norm_axis == 'celltype' else 1

    data = _get_info_from_sample(adata=adata, sample_id=None, key='niche')
    sample = data['sample'].tolist()
    celltype = data['celltype'].tolist()
    sample_uni = np.unique(sample)
    celltype_uni = np.unique(celltype)

    dict_clu = {clu: i for i, clu in enumerate(celltype_uni)}
    dict_sam = {clu: i for i, clu in enumerate(sample_uni)}
    mat_sample_cluster = np.zeros((len(sample_uni), len(celltype_uni)))

    for index in range(len(sample)):
        mat_sample_cluster[dict_sam[sample[index]], dict_clu[celltype[index]]] += 1

    fig, axes = _heatmap_with_dendrogram_and_bar(
        mat_sample_cluster,
        data_count=data,
        x_label='celltype',
        y_label='sample',
        x_map=dict_clu,
        y_map=dict_sam,
        x_dendrogram=False,
        norm_axis=norm_axis,
        method=norm_method,
        x_bar=celltype_bar,
        y_bar=sample_bar,
        cmap=cmap,
        title=title,
        figsize=figsize,
        dpi=dpi,
        xbar_kwags=celltype_bar_kwargs,
        ybar_kwags=sample_bar_kwargs,
        **kwargs
)

    if save:
        fig.savefig(save)
    if show:
        fig.show()
    return fig, axes


def show_celltype_niche_heatmap(
        adata: anndata.AnnData,
        norm_method: Literal['normalization', 'z_score', 'proportion'] = 'normalization',
        niche_bar: bool =True,
        niche_bar_kwargs: Mapping[str, Any] = MappingProxyType({}),
        title: Optional[str] = None,
        figsize: Optional[Tuple[float, float]] = None,
        dpi: int = 100,
        cmap: LinearSegmentedColormap = None,
        show: bool = True,
        save: Union[str, PathLike, None] = None,
        **kwargs,
) -> Optional[Tuple[Figure, dict]]:
    """
    The heatmap shows the cell composition of each niche,
    and the bar plot graph shows the number of each niche in all samples.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    niche_bar : bool
        Whether to show a bar plot of the cell number of each niche.
    niche_bar_kwargs : Any
        Other parameters in sns.countplot() of cell type's bar plot.
    norm_method : Literal['normalization', 'z_score', 'proportion']
        Methods of standardization
        'normalization': x^ = (x - min(x)) / (max(x) - min(x))
        'z_score': x^ = (x - mean(x)) / std(x)
        'proportion': x^ = x / sum(x)
    figsize : Tuple[float, float], optional
        (Width, height) of the figure.
    dpi : float, optional
        The resolution of the figure.
    title : str, optional
        The title of shown figure.
    show : bool
        Show the plot, do not return axis.
    save : Union[str, PathLike], optional
        The path where the image is stored.
    cmap
        Color map to use for continous variables.
    kwargs : Any
        Other params of sns.heatmap()

    Returns
    -------
        If show==False, return Tuple[Figure, list[Axes]]

    """
    if niche_bar_kwargs is None:
        niche_bar_kwargs = {}

    if title is None:
        title = 'Celltype-C_niche'

    data = _get_info_from_sample(adata=adata, sample_id=None, key='niche')
    data = data.fillna(0)
    niche = data.columns[:-4]
    drop_col = data.columns[-4:]
    niche_num = len(np.unique(data['C_niche'].tolist()))

    mat_niche = []
    for i in range(niche_num):
        data_use = data[data['C_niche'] == i]
        for j in drop_col:
            data_use = data_use.drop(j, axis=1)
        mat_use = data_use.values
        mat_niche.append(np.array(np.sum(mat_use, axis=0))/mat_use.shape[0])
    mat_niche = np.array(mat_niche)
    pd_data = pd.DataFrame(mat_niche, index=range(niche_num), columns=niche)
    dict_clu = {clu: i for i, clu in enumerate(pd_data.columns)}
    dict_niche = {clu: i for i, clu in enumerate(pd_data.index)}
    fig, axes = _heatmap_with_dendrogram_and_bar(
        pd_data.values,
        data_count=data,
        x_label='celltype',
        y_label='C_niche',
        x_map=dict_clu,
        y_map=dict_niche,
        x_bar=False,
        y_bar=niche_bar,
        x_dendrogram=False,
        method=norm_method,
        norm_axis=1,
        cmap=cmap,
        title=title,
        figsize=figsize,
        dpi=dpi,
        ybar_kwags=niche_bar_kwargs,
        **kwargs
)

    if save:
        fig.savefig(save)
    if show:
        fig.show()
    return fig, axes


def show_niche_sample_heatmap(
        adata: anndata.AnnData,
        norm_axis: Literal['niche', 'sample', None] = None,
        norm_method: Literal['normalization', 'z_score', 'proportion'] = 'normalization',
        niche_bar: bool = True,
        sample_bar: bool = True,
        niche_bar_kwargs: Any = None,
        sample_bar_kwargs: Any = None,
        title: Optional[str] = None,
        figsize: Optional[Tuple[float, float]] = None,
        dpi: int = 100,
        cmap: LinearSegmentedColormap = None,
        show: bool = True,
        save: Union[str, PathLike, None] = None,
        **kwargs
) -> Optional[Tuple[Figure, dict]]:
    """
    The heatmap shows the niche composition in each sample, and the bar plot shows the total number of each niche and
    the total number of cells in each sample, respectively.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    norm_axis : Literal['celltype', 'sample']
        The direction of standardization.
        'celltype': Standardization was performed for each cell type.
        'sample': Standardization was performed for each sample.
        None: No standardization.
    niche_bar : bool
        Whether to show a bar plot of the cell number of each niche.
    sample_bar : bool
        Whether to show a bar plot of the cell number of each sample.
    niche_bar_kwargs
        Other parameters in sns.countplot() of cell niche's bar plot.
    sample_bar_kwargs
        Other parameters in sns.countplot() of sample's bar plot.
    norm_method : Literal['normalization', 'z_score', 'proportion']
        Methods of standardization
        'normalization': x^ = (x - min(x)) / (max(x) - min(x))
        'z_score': x^ = (x - mean(x)) / std(x)
        'proportion': x^ = x / sum(x)
    figsize : Tuple[float, float], optional
        (Width, height) of the figure.
    dpi : float, optional
        The resolution of the figure.
    title : str, optional
        The title of shown figure.
    show : bool
        Show the plot, do not return axis.
    save : Union[str, PathLike], optional
        The path where the image is stored.
    cmap
        Color map to use for continous variables.
    kwargs : Any
        Other params of sns.heatmap()

    Returns
    -------
        If show==False, return Tuple[Figure, list[Axes]]

    """
    if niche_bar_kwargs is None:
        niche_bar_kwargs = {}
    if sample_bar_kwargs is None:
        sample_bar_kwargs = {}

    if title is None:
        title = 'C_niche-Sample'

    data = _get_info_from_sample(adata=adata, sample_id=None, key='niche')
    norm_axis = 0 if norm_axis == 'niche' else 1

    sample = data['sample'].tolist()
    niche = data['C_niche'].tolist()
    sample_uni = np.unique(sample)
    niche_uni = np.unique(niche)

    dict_niche = {clu: i for i, clu in enumerate(np.unique(niche))}
    dict_sam = {clu: i for i, clu in enumerate(np.unique(sample))}

    mat_sample_niche = np.zeros((len(sample_uni), len(niche_uni)))

    for index in range(len(sample)):
        mat_sample_niche[dict_sam[sample[index]], dict_niche[niche[index]]] += 1

    fig, axes = _heatmap_with_dendrogram_and_bar(
        mat_sample_niche,
        data_count=data,
        x_label='C_niche',
        y_label='sample',
        x_map=dict_niche,
        y_map=dict_sam,
        norm_axis=norm_axis,
        method=norm_method,
        x_bar=niche_bar,
        y_bar=sample_bar,
        cmap=cmap,
        title=title,
        figsize=figsize,
        dpi=dpi,
        xbar_kwags=niche_bar_kwargs,
        ybar_kwags=sample_bar_kwargs,
        **kwargs
)

    if save:
        fig.savefig(save)
    if show:
        fig.show()
    return fig, axes


def show_niche_environment(
        adata: anndata.AnnData,
        niche: Union[int, str],
        celltype_key: str = 'clusters',
        niche_key: str = 'C_niche',
        sample_key: str = 'sample',
        sample_id: Union[str, int] = None,
        scale: Union[str, float] = 'hires',
        palette: Union[Sequence[str], ListedColormap] = None,
        title: Optional[str] = None,
        ax: Optional[Axes] = None,
        sig_niche_line_color: str = '#000000',
        other_niche_line_color: str = '#999999',
        spatial_in_obsm: str = 'spatial',
        show: bool = True,
        save: Union[str, PathLike, None] = None,
        **kwargs
) -> Optional[Axes]:
    """
    Show the environment composition of a niche.
    The cell boundaries belonging to this niche will be highlighted (the color is regulated by the sig_niche_line_color
    parameter), and the cell boundaries' color of other cells are regulated by the other_niche_line_color parameter.
    Cells labeled with cell type color indicate participation in this niche calculation.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    niche

    celltype_key : str
        The label of cell type in adata.obs.
    niche_key : str
        The label of niche in adata.obs.
    sample_key : str
        The label of sample in adata.obs.
    sample : Union[str, int]
        Sample number to be shown.
    scale : Union[str, float]
        Scale used in subsequent analyses. If it's Visium data it can also be HE image labels (hires or lower).
        Most of the time you don't need to change this.
    color

    title : str
        The title of this figure.
    sig_niche_line_color : str
        The color of the outline of cells in selected niche.
    other_niche_line_color : str
        The color of the outline of cells in other niches.
    spatial_in_obsm : str
        Keyword of coordinate information in obsm.
    show : bool
        Show the plot, do not return axis.
    save : Union[str, PathLike]
        The path where the image is stored.
    kwargs : Any
        Other params of voronoi_plot_2d()

    Returns
    -------
        If show==False, return Axes.

    """
    from scipy.spatial import Voronoi, voronoi_plot_2d
    from collections import OrderedDict

    adata = adata.copy()
    if sample_id is not None:
        try:
            adata = adata[adata.obs[sample_key] == sample_id, :].copy()
        except ValueError:
            raise ValueError(f'The {sample_key} was not found in adata.obs.columns')

    indices = _get_info_from_sample(adata, sample_id=sample_id, key='indices')
    adata.obs['attention_label'] = None

    for index in range(adata.obs.shape[0]):
        edges = indices[index]
        index_of_niche = adata.obs.index[index]
        if adata.obs.loc[index_of_niche, niche_key] == niche:
            adata.obs.loc[index_of_niche, 'attention_label'] = 'center'
            for nei in edges:
                index_of_neigh = adata.obs.index[nei]
                if index_of_neigh != index_of_niche and adata.obs.loc[index_of_neigh, 'attention_label'] != 'center':
                    adata.obs.loc[index_of_neigh, 'attention_label'] = 'neighbor'

    if type(scale) != float:
        scale = _scale(adata, None, scale)

    points = adata.obsm[spatial_in_obsm]*scale
    cluster_list = adata.obs[celltype_key]

    cluster_unique = cluster_list.unique()
    cluster_unique = sorted(cluster_unique)
    attention = adata.obs['attention_label'].tolist()

    if not isinstance(palette, dict):
        palette = _get_palette(list(cluster_unique), sort_order=True, palette=palette)
    else:
        palette = palette

    points_color = [palette[index] for index in cluster_list]

    vor = Voronoi(points)

    if ax is None:
        fig = plt.figure(figsize=(12, 8))
        ax = fig.add_axes([0.1, 0.1, 0.5, 0.75])

    voronoi_plot_2d(vor,
                    show_vertices=False,
                    point_size=0,
                    line_colors=other_niche_line_color,
                    ax=ax,
                    **kwargs
                    )

    for r in range(len(vor.point_region)):
        region = vor.regions[vor.point_region[r]]
        if not -1 in region:
            polygon = [vor.vertices[i] for i in region]
            if attention[r] == 'center':
                x = [p[0] for p in polygon]
                y = [p[1] for p in polygon]
                x.append(polygon[0][0])
                y.append(polygon[0][1])
                plt.plot(x, y, color=sig_niche_line_color, linewidth=1.0)
                plt.fill(*zip(*polygon), color=points_color[r], label=cluster_list[r])
            elif attention[r] == 'neighbor':
                plt.fill(*zip(*polygon), color=points_color[r], label=cluster_list[r])
            else:
                pass
    handles, labels = plt.gca().get_legend_handles_labels()
    by_label = OrderedDict(zip(labels, handles))
    new_label = {}
    for key in sorted(by_label):
        new_label[key] = by_label[key]
    plt.legend(new_label.values(), new_label.keys(), bbox_to_anchor=(1.02, 0.8), loc='upper left')
    plt.title(title)

    if save:
        plt.savefig(save)
    if show:
        plt.show()
    else:
        return ax


import copy
import numpy as np
import scanpy as sc
import cv2 as cv
from typing import Optional, Union
from .utils import _assert_variable_in_list
from ..utils import _check_adata_type, _add_info_from_sample, _scale


class Spatialimg(object):
    """
    Description:
    This class's function is that clustering points can be automatically divided into spatial regions
    """

    def __init__(self,
                 adata: sc.AnnData,
                 clusters: Union[list, str, int],
                 KSize: int,
                 cluster_key: Optional[str] = 'cluster_domain',
                 k_blur: Optional[int] = 7,
                 scale: Union[float, str] = 'hires',
                 eliminate_hole: bool = False,
                 remove_small_objects: bool = False,
                 minsize: int = 1000,
                 connectivity: int = 1,
                 ):

        assert cluster_key in adata.obs.columns, "Clustering must be specified before mask generation."
        assert KSize % 2 == 1, "KSize has to be odd number"

        self.adata = adata
        self.adata_obs = copy.deepcopy(adata.obs)
        self.cluster_label = cluster_key
        self.scale = scale

        print(adata.obsm['spatial'].max(axis=0))
        self.heigh = adata.obsm['spatial'].max(axis=0)[1] * self.scale
        self.width = adata.obsm['spatial'].max(axis=0)[0] * self.scale

        crd = self.adata.obsm['spatial'] * float(self.scale)
        self.adata_obs['imagerow'] = crd[:, 1].astype('int64').tolist()
        self.adata_obs['imagecol'] = crd[:, 0].astype('int64').tolist()

        if type(clusters) in [int, str]:
            clusters = [clusters]
        self.clusters = clusters
        self.KSize = (KSize, KSize)
        self.k_blur = k_blur
        self.eliminate_hole = eliminate_hole
        self.remove_small_objects = remove_small_objects
        if remove_small_objects:
            self.minsize = minsize
            self.connectivity = connectivity

    @property
    def Spot_to_Mask(self,
                     ) -> np.ndarray:

        """
        Get mask from spatial domain spot in ST

        Returns
        -------
        numpy.ndarray, mask of the selected domain
        """
        self.img_mask = self.__Preprocessing()
        self.img_mask = self.__Deliate_and_erode()
        if self.eliminate_hole:
            self.img_mask = self.__Eliminate_hole()
        if self.remove_small_objects:
            self.img_mask = self.__Remove_small_objects()

        return self.img_mask

    def __Preprocessing(self):

        """
        Preprocessing function

        :return:
        None
        """
        import math

        for cluster in self.clusters:
            if cluster not in self.adata_obs[self.cluster_label].unique():
                raise ValueError('Input clusters must belong to existing clusters. Checking your input value')

        self.adata_obs.loc[:, 'selected'] = False
        for barcode in self.adata_obs.index:
            if self.adata_obs.loc[barcode, self.cluster_label] in self.clusters:
                self.adata_obs.loc[barcode, 'selected'] = True

        self.adata_obs = self.adata_obs[self.adata_obs['selected']]
        self.mask_data = np.zeros((int(self.heigh * 1.1), int(self.width * 1.1)))

        for index, row in self.adata_obs.iterrows():
            self.mask_data[
                math.floor(row['imagerow']), math.floor(
                    row['imagecol'])] = 255
        kernel = np.ones(self.KSize, dtype=np.uint8)
        imgDilate = cv.dilate(self.mask_data, kernel=kernel).astype('uint8')
        self.mask_data = imgDilate
        return imgDilate

    def __Deliate_and_erode(self,
                            ) -> np.ndarray:

        """
        Function of dilation and erosion

        Returns
        -------
        numpy.ndarray, mask
        """
        kernel = np.ones(self.KSize, dtype=np.uint8)
        imgErode = cv.erode(self.mask_data, kernel=kernel)
        imgErode = imgErode.astype('uint8')
        # imgErode = imgDilate.astype('uint8')

        return imgErode

    def __Eliminate_hole(self,
                         ) -> np.ndarray:

        """
        Fynction that eliminate holes

        Returns
        -------
        numpy.ndarray, mask
        """
        mask = self.img_mask.astype('uint8')
        imgBin = cv.bitwise_not(mask)
        marker = np.zeros_like(imgBin, dtype=np.uint8)
        marker[0, :] = 255
        marker[-1, :] = 255
        marker[:, 0] = 255
        marker[:, -1] = 255

        element = cv.getStructuringElement(shape=cv.MORPH_CROSS, ksize=(5, 5))
        while True:
            marker_pre = marker
            dilation = cv.dilate(marker, kernel=element)
            dilation = dilation.astype('uint8')
            marker = np.min((dilation, imgBin), axis=0)
            if (marker_pre == marker).all():
                break
        imgRebuild = cv.bitwise_not(marker)
        print(imgRebuild)

        return imgRebuild

    def __Remove_small_objects(self,
                               ) -> np.ndarray:

        """
        Function that remove small connected components

        Returns
        -------
        numpy.ndarray, mask
        """

        from skimage import morphology
        imgBlur = cv.blur(self.img_mask, (self.k_blur, self.k_blur))
        ret, imgBlur = cv.threshold(imgBlur, 100, 255, cv.THRESH_BINARY)
        imgBlur = np.array(imgBlur, dtype=bool)
        imgRemove = morphology.remove_small_objects(imgBlur, min_size=self.connectivity, connectivity=self.connectivity)
        imgRemove = imgRemove.astype('uint8') * 255
        return imgRemove


def get_mask_from_domain(
        adata: sc.AnnData,
        clusters: Union[list, str, int],
        KSize: int,
        cluster_key: Optional[str] = 'domain',
        k_blur: Optional[int] = 7,
        scale: Union[str, float] = 'hires',
        eliminate_hole: bool = False,
        remove_small_objects: bool = False,
        minsize: int = 1000,
        connectivity: int = 1,
        inplace: bool = True,
) -> np.ndarray:
    """
    A mask image is generated according to the selected category, and the mask can be processed using morphological methods
    such as dilation and erosion, removal of holes, and removal of small connected components

    The size of mask is based on adata.obsm['spatial']

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    clusters
        Making mask in this clusters
    KSize
        The size of convolution kernel in function dilate
    cluster_key
        Label to which the param 'clusters' belongs
    k_blur
        The size of convolution kernel in function cv2.blur()
    scale
        The spatial scale used by the operation
    eliminate_hole
        If True, use cv.bitwise_not() to eliminate hole
    remove_small_objects
         If True, use morphology.remove_small_objects() to remove small domains
    minsize
        The shortest perimeter of removed domains. Used during
        labelling if `RemoveSmallObjects` is True.
    connectivity
        The connectivity defining the neighborhood of a pixel. Used during
        labelling if `RemoveSmallObjects` is True.
        1 is  four neighborhood, 2 is eight neighborhood.
    inplace : bool, optional
        If True, Modify directly in the original adata.

    Returns
    -------
    numpy.ndarray, mask of the selected domain
    """
    adata = _check_adata_type(adata, 'spatial', inplace)

    if type(scale) != float:
        scale = _scale(adata, None, scale)

    New_Mask = Spatialimg(adata=adata,
                          clusters=clusters,
                          KSize=KSize,
                          cluster_key=cluster_key,
                          k_blur=k_blur,
                          scale=scale,
                          eliminate_hole=eliminate_hole,
                          remove_small_objects=remove_small_objects,
                          minsize=minsize,
                          connectivity=connectivity
                          )

    Mask = New_Mask.Spot_to_Mask
    _add_info_from_sample(adata, sample_id=None, keys='mask', add=Mask)

    return Mask



import anndata as ad
import matplotlib.pyplot as plt
import numpy as np
import copy
import pandas as pd
from typing import Union, Optional, Tuple, Literal, Any
from os import PathLike
from ..utils import _get_info_from_sample
import matplotlib
import matplotlib.patches as patches
from matplotlib.axes import Axes
from matplotlib.figure import Figure

__all__ = ['show_neighborhood_analysis', 'show_infiltration_analysis']


def show_neighborhood_analysis(
        adata: ad.AnnData,
        method: Literal['excluded', 'include'] = 'excluded',
        sample_id: Union[int, str] = None,
        figsize: Optional[Tuple[float, float]] = (8, 8),
        dpi: int = 100,
        title: Optional[str] = None,
        vmin: Optional[float] = None,
        vmax: Optional[float] = None,
        cmap: Any = None,
        show: bool = True,
        save: Union[str, PathLike, None] = None,
        ax: Optional[Axes] = None,
        **kwargs: Any
) -> Optional[Axes]:
    """
    Use heatmap to display the results of neighbor analysis.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    method : str, optional
        'included': Z-scores of edges between two cell types were counted directly after randomization.
        'excluded': After randomization, remove self-connected edges between cells of the same type and calculate the z-score of edges between two cell types.
    sample_id : Union[int, str], optional
        The sample number which to be shown.
    figsize : Tuple[float, float], optional
        (Width, height) of the figure.
    dpi : float, optional
        The resolution of the figure.
    title : str, optional
        The title of shown figure.
    vmin : float, optional
        Lower end of scale bar.
    vmax : float, optional
        Upper end of scale bar.
    cmap : Any, optional
        Color map to use for continuous variables.
    show : bool, optional
        Show the plot, do not return axis.
    save : Union[str, PathLike], optional
        The path where the image is stored.
    ax : Axes, optional
        A matplotlib axes object.
    kwargs : Any
        Other params of sns.heatmap().

    Returns
    -------
    Optional[Axes]
        The matplotlib axes object if `show` is False, otherwise None.
    """
    import seaborn as sns

    if title is None:
        title = 'Neighborhood analysis'

    if ax is None:
        fig, ax = plt.subplots(constrained_layout=True, dpi=dpi, figsize=figsize)

    zscore_df = _get_info_from_sample(adata, sample_id=sample_id, key=method + '_score')
    label = zscore_df.index
    zscore = zscore_df.values

    sns.heatmap(zscore,
                cmap=cmap,
                xticklabels=label,
                yticklabels=label,
                fmt='d',
                cbar=True,
                vmin=vmin,
                vmax=vmax,
                ax=ax,
                **kwargs)

    ax.set_title(title, fontsize=20)

    if save:
        plt.savefig(save)
    if show:
        plt.show()
    else:
        return ax


def show_infiltration_analysis(
        adata: ad.AnnData,
        parenchyma: str,
        nonparenchyma: str,
        min_nonparenchyma: Optional[int] = None,
        sample: Union[list, None] = None,
        figsize: Optional[Tuple[float, float]] = None,
        dpi: int = 100,
        color_num: str = 'red',
        color_score: str = 'blue',
        grid: bool  = True,
        show: bool = True,
        save: Union[str, PathLike, None] = None,
) -> Optional[Tuple[Axes, Axes]]:
    """
    Use barplot to display infiltration scores and the number of non-parenchymal cells.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    parenchyma : str
        The cluster name of parenchyma cell.
    nonparenchyma : str
        The cluster name of non-parenchyma cell.
    min_nonparenchyma

    sample : Union[list, None]
        Samples that need to be presented with infiltration scores.
        By default, all samples for which the infiltration score was calculated are displayed.
    figsize : Tuple[Any, Any]
        The size of format figure.(width, height).
    dpi : int
        Dots per inch values for the output.
    color_num : str
        The color of cells number of non-parenchyma.
    color_score : str
        The color of infiltration score.
    grid : bool
        Whether to add grid or not.
    show : bool
        Show the plot, do not return axis.
    save : Union[str, PathLike]
        The path where the image is stored.

    Returns
    -------
        If show==False, return a tuple of two Axes. ax_left is the barplot Axes of nonparenchyma number,
            and ax_right is the barplot Axes of infiltration score.

    References
    -------
        Nicolas P. Rougier. Scientific Visualization: Python + Matplotlib. Nicolas P. Rougier. , 2021, 978-2-
            9579901-0-8. ffhal-03427242
    """
    data = pd.DataFrame(columns=['sample', 'infiltration score', 'non-number'])

    if sample is None:
        sample = _get_info_from_sample(adata, sample_id=None, key='infiltration_sample')
    for sample_id in sample:
        infiltration = _get_info_from_sample(adata, sample_id=sample_id, key='infiltration_analysis')
        number = _get_info_from_sample(adata, sample_id=sample_id, key='cluster_number')

        data.loc[len(data.index)] = [sample_id, infiltration.loc[parenchyma, nonparenchyma], number[nonparenchyma]]

    max_score = np.floor(max(data['infiltration score']) * 1.2) + 1
    max_number = np.floor(max(data['non-number']) * 1.2) + 1

    data1 = copy.deepcopy(data)
    data1 = data1[data1['non-number'] >= min_nonparenchyma]
    # data1 = data1.sort_values(by='min score', ascending=False)

    right = np.array(data1['infiltration score'].tolist())
    left = np.array(data1['non-number'].tolist())
    sample = data1['sample'].tolist()

    matplotlib.rc("axes", facecolor="white")
    matplotlib.rc("figure.subplot", wspace=0.65)
    matplotlib.rc("grid", color="white")
    matplotlib.rc("grid", linewidth=1)

    fig = plt.figure(figsize=figsize, facecolor="white", dpi=dpi)

    ax_left = plt.subplot(121)

    ax_left.spines["left"].set_color("none")
    ax_left.spines["right"].set_zorder(10)
    ax_left.spines["bottom"].set_color("none")
    ax_left.xaxis.set_ticks_position("top")
    ax_left.yaxis.set_ticks_position("right")
    ax_left.spines["top"].set_position(("data", len(sample) + 0.25))
    ax_left.spines["top"].set_color("w")

    plt.xlim(max_number, 0)
    plt.ylim(0, len(sample))

    plt.xticks([max_number, max_number//2, 0], [str(max_number), str(max_number//2), nonparenchyma])
    ax_left.get_xticklabels()[-1].set_weight("bold")
    ax_left.get_xticklines()[-1].set_markeredgewidth(0)
    for label in ax_left.get_xticklabels():
        label.set_fontsize(10)
    plt.yticks([])

    # Plot data
    for i in range(len(left)):
        H, h = 0.8, 0.55
        value = left[i]
        p = patches.Rectangle(
            (0, i + (1 - H) / 2.0),
            value,
            H,
            fill=True,
            transform=ax_left.transData,
            lw=0,
            facecolor=color_num,
            alpha=0.6,
        )
        ax_left.add_patch(p)

    if grid:
        ax_left.grid()

    ax_right = plt.subplot(122, sharey=ax_left)

    ax_right.spines["right"].set_color("none")
    ax_right.spines["left"].set_zorder(10)
    ax_right.spines["bottom"].set_color("none")
    ax_right.xaxis.set_ticks_position("top")
    ax_right.yaxis.set_ticks_position("left")
    ax_right.spines["top"].set_position(("data", len(sample) + 0.25))
    ax_right.spines["top"].set_color("w")

    plt.xlim(0, 6)
    plt.ylim(0, len(sample))

    plt.xticks(
        [0, max_score//2, max_score],
        ["Infiltration", str(max_score//2), str(max_score)],
    )
    ax_right.get_xticklabels()[0].set_weight("bold")
    for label in ax_right.get_xticklabels():
        label.set_fontsize(10)
    ax_right.get_xticklines()[1].set_markeredgewidth(0)
    plt.yticks([])

    for i in range(len(right)):
        H, h = 0.8, 0.55
        value = right[i]
        p = patches.Rectangle(
            (0, i + (1 - H) / 2.0),
            value,
            H,
            fill=True,
            transform=ax_right.transData,
            lw=0,
            facecolor=color_score,
            alpha=0.6,
        )
        ax_right.add_patch(p)

    if grid:
        ax_right.grid()

    for i in range(len(sample)):
        x1, y1 = ax_left.transData.transform_point((0, i + 0.5))
        x2, y2 = ax_right.transData.transform_point((0, i + 0.5))
        x, y = fig.transFigure.inverted().transform_point(((x1 + x2) / 2, y1))
        sample_id = sample[i]
        plt.text(
            x,
            y,
            sample_id,
            transform=fig.transFigure,
            fontsize=10,
            horizontalalignment="center",
            verticalalignment="center",
        )
    # fig.tight_layout()

    axes = (ax_left, ax_right)

    if save:
        fig.savefig(save)
    if show:
        fig.show()
    else:
        return axes


import copy
import pandas as pd
import scanpy as sc
import numpy as np
from typing import Optional, Union
from ..utils import _scale, _add_info_from_sample, _get_info_from_sample, _check_adata_type
from typing import Optional, Literal
import anndata
import logging as logg
import os

__all__ = ["domain_from_unsupervised", "domain_from_local_moran", "global_moran", "cal_aucell"]


class _SpatialDomain(object):

    def __init__(self,
                 adata: sc.AnnData,
                 domain_method: str,
                 inplace: bool = True,
                 ):

        if domain_method == 'stagate':
            self.domain_emb = 'STAGATE'
        elif domain_method == 'graphst':
            self.domain_emb = 'emb'
        elif domain_method == 'scanit':
            self.domain_emb = 'X_scanit'

        if inplace:
            self.adata = adata
        else:
            self.adata = copy.deepcopy(self.adata)

    def get_stagate_domain(self,
                           graph_model='Radius',
                           rad_cutoff=None,
                           k_cutoff=None,
                           **kwargs
                           ):

        from .other_package_without_pip import STAGATE_pyG as STAGATE

        adata = self.adata

        STAGATE.Cal_Spatial_Net(adata, model=graph_model, k_cutoff=k_cutoff, rad_cutoff=rad_cutoff)
        # STAGATE_pyG.Stats_Spatial_Net(adata)

        STAGATE.train_STAGATE(adata, **kwargs)

        return adata

    def get_graphST_domain(self,
                           **kwargs
                           ):

        from .other_package_without_pip.GraphST import GraphST
        from sklearn.decomposition import PCA

        adata = self.adata
        model = GraphST.GraphST(adata, **kwargs)

        # train model
        adata = model.train()

        pca = PCA(n_components=20, random_state=42)
        embedding = pca.fit_transform(adata.obsm[self.domain_emb].copy())
        adata.obsm['emb_pca'] = embedding
        self.domain_emb = 'emb_pca'
        self.adata = adata

        return adata

    def get_scanit_domain(self,
                          graph_model='knn',
                          alpha_n_layer=1,
                          k_cutoff=10,
                          **kwargs
                          ):
        from .other_package_without_pip import scanit

        adata = self.adata
        adata.X = adata.X.toarray()
        scanit.tl.spatial_graph(adata, method=graph_model, knn_n_neighbors=k_cutoff, alpha_n_layer=alpha_n_layer)
        scanit.tl.spatial_representation(adata, **kwargs)
        self.adata = adata

        return adata

    def mclust_R(self,
                 num_cluster: int,
                 key_added: str = 'domain',
                 modelNames='EEE',
                 random_seed: int = 2020,
                 ):

        import rpy2.robjects as robjects
        import rpy2.robjects.numpy2ri
        adata = self.adata

        np.random.seed(random_seed)
        robjects.r.library("mclust")
        rpy2.robjects.numpy2ri.activate()
        r_random_seed = robjects.r['set.seed']
        r_random_seed(random_seed)
        rmclust = robjects.r['Mclust']

        res = rmclust(rpy2.robjects.numpy2ri.numpy2rpy(adata.obsm[self.domain_emb]), num_cluster, modelNames)
        mclust_res = np.array(res[-2])

        adata.obs[key_added] = mclust_res
        adata.obs[key_added] = adata.obs[key_added].astype('int')
        adata.obs[key_added] = adata.obs[key_added].astype('category')

        return adata

    def louvain(self,
                resolution: float = 0.5,
                key_added: str = 'domain',
                ):

        adata = self.adata
        sc.pp.neighbors(adata, use_rep=self.domain_emb)
        sc.tl.umap(adata)
        # louvain
        sc.tl.louvain(adata, resolution=resolution, key_added=key_added)

        return adata


def domain_from_unsupervised(
        adata: anndata.AnnData,
        domain_method: Literal['stagate', 'graphst', 'scanit'] = 'stagate',
        graph_model: str = None,
        k_cutoff: Optional[int] = None,
        rad_cutoff: Optional[float] = None,
        alpha_n_layer: Optional[int] = None,
        cluster_method: Literal['m_clust', 'louvain'] = 'm_clust',
        cluster_number: int = 10,
        cluster_key: str = 'domain',
        random_seed: int = 2020,
        resolution_louvain: float = 0.5,
        spatial_in_obsm: str = 'spatial',
        inplace: bool = True,
        **kwargs
) -> anndata.AnnData:
    """
    Generate spatial domains using unsupervised learning methods.
    This function supports multiple spatial domain identification algorithms, including STAGATE, GraphST, and ScanIT,
    and provides two clustering methods (mclust and Louvain) to cluster the spatial domains.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.

    domain_method : Literal['stagate', 'graphst', 'scanit'], optional (default: 'stagate')
        The method used to generate spatial domains. Available options are:
        - 'stagate': Use the STAGATE algorithm to generate spatial domains.
        - 'graphst': Use the GraphST algorithm to generate spatial domains.
        - 'scanit': Use the ScanIT algorithm to generate spatial domains.

    graph_model : str, optional (default: None)
        The model used to construct the spatial graph. For STAGATE, options are 'Radius' or 'KNN';
        for ScanIT, options are 'alpha shape' or 'knn'.

    k_cutoff : Optional[int], optional (default: None)
        The number of KNN neighbors used to construct the spatial graph. Only valid when graph_model is 'KNN' or 'knn'.

    rad_cutoff : Optional[float], optional (default: None)
        The radius cutoff used to construct the spatial graph. Only valid when graph_model is 'Radius'.

    alpha_n_layer : Optional[int], optional (default: None)
        The number of alpha layers used in the ScanIT algorithm. Only valid when domain_method is 'scanit'.

    cluster_method : Literal['m_clust', 'louvain'], optional (default: 'm_clust')
        The clustering algorithm used. Available options are:
        - 'm_clust': Use the mclust algorithm for clustering.
        - 'louvain': Use the Louvain algorithm for clustering.

    cluster_number : int, optional (default: 10)
        The number of clusters. Only valid when cluster_method is 'm_clust'.

    cluster_key : str, optional (default: 'domain')
        The key in adata.obs where the clustering results will be stored.

    random_seed : int, optional (default: 2020)
        Random seed for reproducibility.

    resolution_louvain : float, optional (default: 0.5)
        The resolution parameter for the Louvain algorithm. Only valid when cluster_method is 'louvain'.

    spatial_in_obsm : str, optional (default: 'spatial')
        The key in adata.obsm where spatial coordinates are stored.

    inplace : bool, optional (default: True)
        Whether to modify the AnnData object in place. If False, a modified copy is returned.

    **kwargs : dict
        Additional parameters passed to the specific algorithms.


    Returns
    -------
    - :attr:`anndata.AnnData.obs` ``domain`` - The cluster of spatial domain.

    """

    adata = _check_adata_type(adata, spatial_in_obsm, inplace)
    if spatial_in_obsm != 'spatial':
        # stagate only recognize coordinate information from adata.obsm['spatial']
        adata.obsm['spatial'] = adata.obsm[spatial_in_obsm]

    New_SpatialDomain = _SpatialDomain(adata, domain_method=domain_method, inplace=inplace)

    if domain_method == 'stagate':

        if graph_model == None:
            graph_model = 'Radius'
        elif graph_model == 'knn':
            graph_model = 'KNN'

        assert (graph_model in ['Radius', 'KNN']), 'graph_model of STAGATE must in [\'Radius\', \'KNN\']'

        adata = New_SpatialDomain.get_stagate_domain(
            graph_model=graph_model,
            k_cutoff=k_cutoff,
            rad_cutoff=rad_cutoff,
            **kwargs
            )
    elif domain_method == 'graphst':
        adata = New_SpatialDomain.get_graphST_domain(
            **kwargs
            )
    elif domain_method == 'scanit':

        if graph_model == None or graph_model == 'KNN':
            graph_model = 'knn'

        assert (graph_model in ['alpha shape', 'knn']), 'graph_model of scanit must in [\'alpha shape\', \'knn\']'

        adata = New_SpatialDomain.get_scanit_domain(
            graph_model=graph_model,
            alpha_n_layer=alpha_n_layer,
            k_cutoff=k_cutoff,
            **kwargs
            )

    if cluster_method == 'm_clust':
        New_SpatialDomain.mclust_R(num_cluster=cluster_number,
                                   key_added=cluster_key,
                                   random_seed=random_seed,
                                   )
    elif cluster_method == 'louvain':
        New_SpatialDomain.louvain(resolution=resolution_louvain,
                                  key_added=cluster_key
                                  )
    else:
        logg.error(f'{cluster_method} is not in [\'m_clust\', \'louvain\']', exc_info=True)
        raise ValueError()

    return adata


def domain_from_local_moran(
        adata: anndata.AnnData,
        score_key: str,
        moran_label_key: str = 'hotspot',
        k: int = 6,
        pvalue_filter: float = 0.05,
        attribute_filter: Optional[float] = None,
        spatial_lag_filter: Optional[float] = None,
        zscore: bool = True,
        fdr: bool = True,
        scale: Union[str, float] = 'hires',
        spatial_in_obsm: str = 'spatial',
        inplace: bool = True,
) -> anndata.AnnData:
    """
    Using the local moran method generate the spatial domain.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    score_key : str
        The key for the Moran index need to be computed. It could be gene name ,key in adata.obs_index or AUCell name.
    moran_label_key : str
        The key for storing Moran labels in adata.obs.
    k : int, optional
        Number of nearest neighbors in KNN.
    pvalue_filter : float, optional
        Threshold for determining hotspot regions based on the p-value of the local Moran's I statistic.
        A smaller value indicates higher statistical significance, typically set to 0.05 or another significance level.
        For a detailed explanation, see https://pysal.org/esda/generated/esda.Moran_Local.html#esda.Moran_Local
    attribute_filter : float, optional
        Threshold for attribute filtering (AUCell score).
        If provided, it affects how observations are labeled as "Hotspot" or "Not" based on whether the attribute value
        is greater than or equal to this threshold.
        For a detailed explanation, see https://pysal.org/esda/generated/esda.Moran_Local.html#esda.Moran_Local
    spatial_lag_filter : float, optional
        Threshold for the average attribute of neighboring spots.
        If provided, it affects how observations are labeled as "Hotspot" or "Not" based on whether the spatial lag
        value of the local Moran's I statistic is greater than or equal to this threshold.
    zscore : bool, optional
        If True, calculate the z-score of the attribute.
    fdr : bool, optional
        If True, the p-values were corrected for FDR.
    scale : Union[str, float], optional
        scale used in subsequent analyses. If it's Visium data it can also be HE image labels (hires or lower).
        Most of the time you don't need to change this.
    spatial_in_obsm : str, optional
        The key of spatial coordinates in adata.obsm.
    inplace : bool, optional
        If True, Modify directly in the original adata.

    Returns
    -------
    - :attr:`anndata.AnnData.uns` ``['SOAPy']['local_moran']`` - Local Moran information for each spot

    """
    import libpysal
    from esda.moran import Moran_Local
    from libpysal.weights.spatial_lag import lag_spatial
    from pyscenic.binarization import derive_threshold

    adata = _check_adata_type(adata, spatial_in_obsm, inplace)
    if type(scale) != float:
        scale = _scale(adata, None, scale)

    df_pixel = adata.obsm[spatial_in_obsm]
    if isinstance(df_pixel, np.ndarray):
        df_pixel = pd.DataFrame(df_pixel)
    df_pixel = df_pixel.iloc[:, [0, 1]]

    obs = copy.deepcopy(adata.obs)
    obs['x'] = list(df_pixel.iloc[:, 0] * scale)
    obs['y'] = list(df_pixel.iloc[:, 1] * scale)
    coord_df = obs[['y', 'x']]

    try:
        score = _get_info_from_sample(adata, sample_id=None, key='aucell', printf=False)
        score = score[score_key]
    except KeyError:
        if score_key in adata.obs.columns:
            score = adata.obs[score_key]
        else:
            var_names = adata.var_names.tolist()
            index = var_names.index(score_key)
            express = adata.X
            # if type(adata.X) is np.ndarray:
            #     express = adata.X
            # else:
            #     express = adata.X.toarray()
            score = {score_key: express[:, index].tolist()}
            score = pd.DataFrame(score, index=adata.obs.index)

    w = libpysal.weights.KNN(coord_df, k=k)
    moran = Moran_Local(score.astype(np.float64), w)

    if zscore:
        local_moran_df = pd.DataFrame({'Attribute': moran.z,
                                       'Spatial Lag': lag_spatial(moran.w, moran.z),
                                       'Moran_Is': moran.Is,
                                       'Cluster': moran.q,
                                       'P_value': moran.p_sim})
    else:
        local_moran_df = pd.DataFrame({'Attribute': score.tolist(),
                                       'Spatial Lag': lag_spatial(moran.w, score.tolist()),
                                       'Moran_Is': moran.Is,
                                       'Cluster': moran.q,
                                       'P_value': moran.p_sim})

    if fdr:
        from statsmodels.stats.multitest import fdrcorrection
        _, p_adjust = fdrcorrection(local_moran_df['P_value'])
        local_moran_df['P_value'] = p_adjust

    if attribute_filter is None:
        attribute_filter = derive_threshold(local_moran_df, 'Attribute')
    print('attribute_filter:', attribute_filter)
    if spatial_lag_filter is None:
        spatial_lag_filter = derive_threshold(local_moran_df, 'Spatial Lag')
    print('spatial_lag_filter:', spatial_lag_filter)

    local_moran_df['Label'] = [
        'Hotspot' if (local_moran_df.loc[i, 'P_value'] <= pvalue_filter) and (
                    local_moran_df.loc[i, 'Attribute'] >= attribute_filter) and (
                    local_moran_df.loc[i, 'Spatial Lag'] >= spatial_lag_filter) else 'Not' for i in
        range(local_moran_df.shape[0])]

    local_moran_df.index = coord_df.index
    adata.obs['Moran_domain'] = local_moran_df['Label']

    local_moran = {'LocalMoran': local_moran_df,
                   'attribute_filter': attribute_filter,
                   'spatial_lag_filter': spatial_lag_filter,
                   'pvalue_filter': pvalue_filter}

    adata.obs[moran_label_key] = local_moran_df['Label'].tolist()
    _add_info_from_sample(adata, sample_id=None, keys='local_moran', add=local_moran)

    return adata


def global_moran(
        adata: anndata.AnnData,
        score_labels: Union[list, str, None] = None,
        k: int = 6,
        scale: Union[str, float] = 'hires',
        spatial_in_obsm: str = 'spatial',
        inplace: bool = True,
) -> anndata.AnnData:
    """
    The global Moran index is calculated based on selected genes or indicators such as AUCell score.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    score_labels : str
        The label or label lists for the Moran index need to be computed.
        It could be gene name in adata.uns_names or in adata.uns['SOAPy']['aucell'] (the result of cal_aucell()).
    k : int, optional
        Number of nearest neighbors in KNN
    scale : Union[str, float], optional
        The scaling factor for distance scaling. If it's Visium data it can also be HE image labels (hires or lower).
        Most of the time you don't need to change this.
    spatial_in_obsm : str, optional
        Keyword of coordinate information in obsm.
    inplace : bool, optional
        Whether to change the original adata.

    Returns
    -------
    - :attr:`anndata.AnnData.uns` ``['SOAPy']['global_moran']`` - Global Moran information for each keyword

    """
    from tqdm import tqdm
    from esda.moran import Moran
    import libpysal
    from shapely import geometry
    import geopandas as gpd

    adata = _check_adata_type(adata, spatial_in_obsm, inplace)

    if score_labels is None:
        data_df = _get_info_from_sample(adata, sample_id=None, key='aucell')
    else:
        if isinstance(score_labels, str):
            score_labels = [score_labels]
        for score_label in score_labels:
            data_df = pd.DataFrame(index=adata.obs.index)
            try:
                score = _get_info_from_sample(adata, sample_id=None, key='aucell')
                data_df[score_label] = score[score_label].tolist()
            except KeyError:
                var_names = adata.var_names.tolist()
                if score_label in var_names:
                    data_df[score_label] = adata[:, score_label].X.toarray().T[0]
                else:
                    logg.error(f'{score_label} is not in the aucell and gene list, please check. '
                               f'This keyword has been skipped')
                    continue

    if type(scale) != float:
        scale = _scale(adata, None, scale)

    df_pixel = adata.obsm[spatial_in_obsm]
    if isinstance(df_pixel, np.ndarray):
        df_pixel = pd.DataFrame(df_pixel)
    df_pixel = df_pixel.iloc[:, [0, 1]]

    obs = copy.deepcopy(adata.obs)
    obs['x'] = list(df_pixel.iloc[:, 0] * scale)
    obs['y'] = list(df_pixel.iloc[:, 1] * scale)
    coord_df = obs[['y', 'x']]

    coord_df = coord_df.loc[data_df.index, :]
    coord_point = [geometry.Point(coord_df.iloc[i, 0], coord_df.iloc[i, 1]) for i in range(coord_df.shape[0])]

    geo_df = gpd.GeoDataFrame(data=data_df,
                              geometry=coord_point)
    w = libpysal.weights.KNN.from_dataframe(geo_df, k=k)

    # Global Moran index
    geo_mat = geo_df.values
    moran = [Moran(geo_mat[:, i].astype('float64'), w) for i in tqdm(range(geo_df.shape[1] - 1))]
    moran_I = [m.I for m in moran]
    moran_p = [m.p_sim for m in moran]
    global_moran = pd.DataFrame({'Moran_I': moran_I,
                                 'P_value': moran_p})

    global_moran.index = data_df.columns.tolist()[0:data_df.shape[1] - 1]

    _add_info_from_sample(adata, sample_id=None, keys='global_moran', add=global_moran)

    return adata


def cal_aucell(
        adata: anndata.AnnData,
        signatures: dict,
        gene_percent: float = 0.05,
        inplace: bool = True,
        **kwargs
) -> anndata.AnnData:
    """
    AUCell scores were calculated for given gene lists.
    For more information about AUCell, see https://www.nature.com/articles/s41596-020-0336-2

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    signatures : dict
        The name of each gene set and the corresponding gene list.
    gene_percent : float, optional
        The fraction of the ranked genome to take into account for the calculation of the area Under the recovery Curve.
    inplace : bool, optional
        Whether to change the original adata.
    kwargs:
        Other Parameters of pyscenic.aucell.aucell.

    Returns
    -------
        - :attr:`anndata.AnnData.uns` ``['SOAPy']['global_moran']`` - AUCell score of signatures
    """
    from ctxcore.genesig import GeneSignature
    from pyscenic.aucell import aucell

    adata = _check_adata_type(adata, 'spatial', inplace)

    exp_mat=pd.DataFrame(adata.X.toarray(), index=adata.obs_names, columns=adata.var_names)
    # generate GeneSignature list
    gs = [GeneSignature(name=k, gene2weight=v) for (k, v) in signatures.items()]
    gene_thres = pd.Series(np.count_nonzero(exp_mat, axis=1)).quantile(gene_percent)/exp_mat.shape[1]
    aucell_df = aucell(exp_mat, gs, auc_threshold=gene_thres, **kwargs)
    _add_info_from_sample(adata, sample_id=None, keys='aucell', add=aucell_df)

    return adata




import numpy as np
from scipy import sparse
from sklearn.neighbors import kneighbors_graph

def estimate_cutoff_knn(pts, k=10):
    A_knn = kneighbors_graph(pts, n_neighbors=k, mode='distance')
    est_cut = A_knn.sum() / float(A_knn.count_nonzero())
    return est_cut


import copy
from typing import Optional, Union, Literal
import logging as logg
import anndata
from tqdm import tqdm
import anndata as ad
import scanpy as sc
import pandas as pd
import numpy as np
import cv2 as cv
import statsmodels.api as sm
from skmisc.loess import loess
from ..utils import _scale, _add_info_from_sample, _get_info_from_sample, _check_adata_type
import warnings

__all__ = ["wilcoxon_test", "spearman_correlation", "ANOVA", "spatial_tendency", "gene_cluster", "adata_from_mask"]


def polynomial_regression(
        dis_and_express: pd.DataFrame,
        gene_name: str,
        frac: int,
        sd: Optional[int] = None,
        drop_zero: bool = False,
):
    """
    polynomial regression function
    """
    mat_only_gene_express = dis_and_express.loc[:, ['distance', gene_name]].values

    if sd != None:
        y = mat_only_gene_express[:, 1]
        mu = np.mean(y)
        std = np.std(y)
        mat_only_gene_express[(mat_only_gene_express[:, 1] < mu - sd * std), 1] = mu - sd * std
        mat_only_gene_express[(mat_only_gene_express[:, 1] > mu + sd * std), 1] = mu + sd * std

    if drop_zero:
        mat_only_gene_express = mat_only_gene_express[~(mat_only_gene_express[:, 1] == 0)]

    x = mat_only_gene_express[:, 0]
    y = mat_only_gene_express[:, 1]
    ind = x.argsort()
    x = x[ind]
    y = y[ind]
    X = x

    if frac > 1:
        for index in range(frac - 1):
            X = np.column_stack((X, x ** (index + 2)))
    X = sm.add_constant(X)

    model = sm.OLS(y.astype(float), X)
    res = model.fit()
    y_fitted = res.fittedvalues

    y_min = y_fitted.min()
    y_max = y_fitted.max()
    ran = y_max - y_min

    aic = res.aic
    bic = res.bic

    corr = np.corrcoef(x=x, y=y)[0, 1]

    return res.f_pvalue, x, y_fitted, res.params, ran, corr, len(y), aic, bic


def loess_(
        dis_and_express: pd.DataFrame,
        gene_name: str,
        frac_0_1: float,
        sd: Optional[int] = None,
        drop_zero: bool = False,
):
    """
    loess function
    """

    mat_only_gene_express = dis_and_express.loc[:, ['distance', gene_name]].values

    if sd is not None:
        y = mat_only_gene_express[:, 1]
        mu = np.mean(y)
        std = np.std(y)
        mat_only_gene_express[(mat_only_gene_express[:, 1] < mu - sd * std), 1] = mu - sd * std
        mat_only_gene_express[(mat_only_gene_express[:, 1] > mu + sd * std), 1] = mu + sd * std

    if drop_zero:
        mat_only_gene_express = mat_only_gene_express[~(mat_only_gene_express[:, 1] == 0)]

    x = mat_only_gene_express[:, 0]
    y = mat_only_gene_express[:, 1]

    # x = np.array(x)
    # y = np.array(y)
    ind = x.argsort()
    y = y[ind]
    x = x[ind]

    l = loess(x, y, span=frac_0_1)
    l.fit()
    pred = l.predict(x, stderror=True)
    yest = pred.values

    yBar = np.mean(y)
    SSE = 0
    SST = 0
    for i in range(0, len(x)):
        SSE += (yest[i] - y[i]) ** 2
        SST += (y[i] - yBar) ** 2

    r_sq = round(1 - SSE / SST, 3)
    corr = np.corrcoef(x=x, y=y)[0, 1]

    y_min = yest.min()
    y_max = yest.max()
    ran = y_max - y_min

    return x, yest, r_sq, corr, ran, len(y)


class SpatialTendency(object):

    def __init__(self,
                 adata: sc.AnnData,
                 gene_name: Union[str, list],
                 radius: Optional[float] = None,
                 scale: Union[float, str] = None,
                 clusters: Union[str, int, list] = 'all',
                 cluster_key: Optional[str] = None,
                 spatial_in_obsm: str = 'spatial',
                 ) -> None:
        """
        initialization
        """

        if clusters != 'all':
            if type(clusters) != list:
                clusters = [clusters]
            self.tendency_selected = [(i in clusters) for i in adata.obs[cluster_key]]

        if type(adata.X) is not np.ndarray:
            df = pd.DataFrame(adata.X.toarray())
        else:
            df = pd.DataFrame(adata.X)

        df.index = adata.obs.index
        df.columns = adata.var.index
        df_pixel = adata.obsm[spatial_in_obsm]
        # df_pixel = df_pixel.iloc[:, [0, 1]]

        if type(scale) != float:
            scale = _scale(adata, None, scale)
        df_pixel = pd.DataFrame(df_pixel, index=adata.obs_names) * scale

        adata.obs['imagerow'] = df_pixel[1].tolist()
        adata.obs['imagecol'] = df_pixel[0].tolist()

        self.adata = adata
        self.scale = str(scale)

        self.express = df
        if clusters != 'all':
            self.express = self.express.loc[self.tendency_selected, :]
        self.coordinate = adata.obs.loc[:, ['imagerow', 'imagecol']]
        if clusters != 'all':
            self.coordinate = self.coordinate.loc[self.tendency_selected, :]
        if type(gene_name) != list:
            gene_name = [gene_name]
        self.gene_name = gene_name
        if radius is None:
            radius = np.inf
        self.radius = radius
        self.inSpots = []

        if radius <= 0:
            raise RuntimeError("Radius should greater than 0")
        self.__structureCheck()

    def __structureCheck(self, ) -> None:
        """
        checking for mismatched data

        """

        shape_express = self.express.shape
        shape_coordinate = self.coordinate.shape

        if len(shape_express) > 2:
            raise RuntimeError("Express should be 2-dimensionality")
        if len(shape_coordinate) > 2:
            raise RuntimeError("Coordinate should be 2-dimensionality")

    def __distance_spot_and_contour(self,
                                    contour: np.ndarray,
                                    location: Optional[str],
                                    ) -> pd.DataFrame:
        """
        The distance from each spot to the mask boundary is calculated, positive outside the mask and negative
        inside the mask
        """

        spot_x = []
        spot_y = []
        dists = []
        exps = []

        for gene_name in self.gene_name:
            exps.append([])
        threshold = self.__find_threshold(contour, self.radius)
        crds = self.coordinate

        for index in crds.index:
            point = (crds.loc[index, 'imagecol'], crds.loc[index, 'imagerow'])
            if point in self.inSpots:
                continue
            if not self.__fliter(point, threshold):
                continue

            point = (float(point[0]), float(point[1]))
            dist = cv.pointPolygonTest(contour, point, True)
            dist = -dist

            if location == 'out':
                if dist < 0 or abs(dist) > self.radius:  ##########
                    if dist < 0:  ###############
                        self.inSpots.append(point)  ##################
                    continue
            if location == 'in':
                if dist > 0 or abs(dist) > self.radius:  ##########
                    continue
            if location == 'all':
                if abs(dist) > self.radius:
                    continue
            for index_gene, gene_name in enumerate(self.gene_name):
                exps[index_gene].append(self.express.at[index, gene_name])
            spot_x.append(point[0])
            spot_y.append(point[1])
            dists.append(dist)

        contour_dis_and_express = {'point_x': spot_x, 'point_y': spot_y, 'distance': dists}

        for index, gene_name in enumerate(self.gene_name):
            contour_dis_and_express[gene_name] = exps[index]

        contour_dis_and_express = pd.DataFrame(contour_dis_and_express)

        return contour_dis_and_express

    def __find_threshold(self,
                         contour: np.ndarray,
                         radius: float,
                         ) -> list:

        left_point = tuple(contour[contour[:, :, 0].argmin()][0])
        right_point = tuple(contour[contour[:, :, 0].argmax()][0])
        top_point = tuple(contour[contour[:, :, 1].argmin()][0])
        bottom_point = tuple(contour[contour[:, :, 1].argmax()][0])

        return [left_point[0] - radius, right_point[0] + radius, top_point[1] - radius, bottom_point[1] + radius]

    def __fliter(self,
                 point: tuple,
                 threshold: list,
                 ) -> bool:
        """
        Pre-filter spots, keeping points that are likely to be within distance
        """
        if point[0] < threshold[0] or point[0] > threshold[1]:
            return False
        if point[1] < threshold[2] or point[1] > threshold[3]:
            return False

        return True

    def get_new_adata(self,
                      mask: np.ndarray,
                      location: Optional[str],
                      radius,
                      ):
        contours, hierarchy = cv.findContours(mask.copy(), cv.RETR_EXTERNAL, cv.CHAIN_APPROX_NONE)

        self.adata.obs['in_mask'] = None

        for contour in contours:
            threshold = self.__find_threshold(contour, self.radius)

            for index in self.adata.obs_names.tolist():
                point = (self.adata.obs.loc[index, 'imagecol'], self.adata.obs.loc[index, 'imagerow'])
                if not self.__fliter(point, threshold):
                    continue

                point = (float(point[0]), float(point[1]))
                marker = cv.pointPolygonTest(contour, point, False)
                marker = -marker
                if self.adata.obs.loc[index, 'in_mask'] is None:
                    self.adata.obs.loc[index, 'in_mask'] = marker
                else:
                    if marker < 0:
                        self.adata.obs.loc[index, 'in_mask'] = marker
                    else:
                        self.adata.obs.loc[index, 'in_mask'] = np.min((self.adata.obs.loc[index, 'in_mask'], marker))

        if radius is None:
            radius = np.inf
        if location == 'in':
            left = -radius
            right = 0
        elif location == 'out':
            left = 0
            right = radius
        else:
            left = -radius
            right = radius

        bdata = self.adata[np.isin(left <= self.adata.obs['in_mask'], True, False), :].copy()
        bdata = bdata[np.isin(bdata.obs['in_mask'] <= right, True, False), :].copy()

        return bdata

    def get_dist_and_express(self,
                             mask: np.ndarray,
                             location: Optional[str] = 'all',
                             ) -> pd.DataFrame:

        contours, hierarchy = cv.findContours(mask.copy(), cv.RETR_EXTERNAL, cv.CHAIN_APPROX_NONE)

        index = 0
        dis_and_express = []
        x_all = []
        y_all = []
        dis_all = []
        express_all = []

        for index_gene, gene_name in enumerate(self.gene_name):
            express_all.append([])

        # Get the data for each spot
        for contour in contours:
            contour_dis_and_express = self.__distance_spot_and_contour(contour, location)
            dis_and_express.append([contour, dis_and_express])
            x_all = x_all + contour_dis_and_express['point_x'].tolist()
            y_all = y_all + contour_dis_and_express['point_y'].tolist()
            dis_all = dis_all + contour_dis_and_express['distance'].tolist()
            for index_gene, gene_name in enumerate(self.gene_name):
                express_all[index_gene] = express_all[index_gene] + contour_dis_and_express[gene_name].tolist()
            index += 1

        # Integrate data from all spots
        dis_and_express_all = {'x': x_all, 'y': y_all, 'distance': dis_all}
        for index_gene, gene_name in enumerate(self.gene_name):
            dis_and_express_all[gene_name] = express_all[index_gene]

        dis_and_express_all = pd.DataFrame(dis_and_express_all)
        dis_and_express_all = self.__remove_duplicates(dis_and_express_all)

        for index_point in dis_and_express_all.index:
            if (dis_and_express_all.loc[index_point, 'x'], dis_and_express_all.loc[index_point, 'y']) in self.inSpots:
                dis_and_express_all.drop(labels=index_point, inplace=True)

        _add_info_from_sample(self.adata, sample_id=None, keys='distance_and_expression', add=dis_and_express_all)
        _add_info_from_sample(self.adata, sample_id=None, keys='contours', add=contours)

        return dis_and_express_all

    def loess(self,
              mask: np.ndarray,
              frac_0_1: float = 0.6666,
              sd: Optional[int] = None,
              location: Optional[str] = 'all',
              drop_zero: bool = False,
              ) -> anndata.AnnData:
        """
        loess regression of expression and distance
        """

        warnings.filterwarnings('ignore')
        dis_and_express = self.get_dist_and_express(mask, location)

        dic_crd = {}
        list_name = []
        list_r_sq_lowess = []
        list_corr_gene_distance = []
        list_range_lowess = []
        list_num_of_spots = []

        mark = 0
        for gene_name in tqdm(self.gene_name):
            xest, yest, r_sq, corr, ran, spot_num = loess_(dis_and_express,
                                                           gene_name,
                                                           frac_0_1,
                                                           sd=sd,
                                                           drop_zero=drop_zero
                                                           )
            if mark == 0:
                dic_crd['Xest'] = xest
                mark = 1
            dic_crd[gene_name] = yest
            list_name.append(gene_name)
            list_r_sq_lowess.append(r_sq)
            list_corr_gene_distance.append(corr)
            list_range_lowess.append(ran)
            list_num_of_spots.append(spot_num)

        dict_gene = {
            'R_square': list_r_sq_lowess,
            'correlation': list_corr_gene_distance,
            'range': list_range_lowess,
            'Spots number': list_num_of_spots
        }

        df_param_loess = pd.DataFrame(dict_gene, index=list_name)

        add_data = {'dic_crd_loess': dic_crd, 'df_param_loess': df_param_loess, 'loess_frac': frac_0_1}
        _add_info_from_sample(self.adata, sample_id=None, keys='loess', add=add_data)

        return self.adata

    def polynomialRegression(self,
                             mask: np.ndarray,
                             frac: int = 4,
                             sd: Optional[int] = None,
                             location: Optional[str] = 'out',
                             drop_zero: bool = False,
                             ) -> anndata.AnnData:
        """
        Polynomial regression of expression and distance
        """

        warnings.filterwarnings('ignore')
        dis_and_express = self.get_dist_and_express(mask, location)

        dic_crd = {}
        list_name = []
        list_p_value_Poly = []
        list_param_Poly = []
        list_range_Poly = []
        list_corr_gene_distance = []
        list_num_of_spots = []
        list_aic = []
        list_bic = []

        mark = 0
        for gene_name in tqdm(self.gene_name):
            f_pvalue, xest, yest, param, ran, corr, spot_num, aic, bic = polynomial_regression(
                dis_and_express=dis_and_express,
                gene_name=gene_name,
                frac=frac,
                sd=sd,
                drop_zero=drop_zero
                )
            if mark == 0:
                dic_crd['Xest'] = xest
                mark = 1
            dic_crd[gene_name] = yest
            list_name.append(gene_name)
            list_p_value_Poly.append(f_pvalue)
            list_param_Poly.append(param)
            list_range_Poly.append(ran)
            list_corr_gene_distance.append(corr)
            list_num_of_spots.append(spot_num)
            list_aic.append(aic)
            list_bic.append(bic)

        dict_PR = {
            'p_value': list_p_value_Poly,
            'param': list_param_Poly,
            'range': list_range_Poly,
            'correlation': list_corr_gene_distance,
            'Spots number': list_num_of_spots,
            'AIC': list_aic,
            'BIC': list_bic
        }

        df_param_poly = pd.DataFrame(dict_PR, index=list_name)

        add_data = {'dic_crd_poly': dic_crd, 'df_param_poly': df_param_poly, 'poly_frac': frac}
        _add_info_from_sample(self.adata, sample_id=None, keys='poly', add=add_data)

        return self.adata

    def __remove_duplicates(self,
                            dis_and_express: pd.DataFrame,
                            ) -> pd.DataFrame:
        """
        removing duplicate spots
        """
        dis_and_express['distance_abs'] = abs(dis_and_express['distance'])
        dis_and_express = dis_and_express.sort_values(by='distance_abs')
        dis_and_express = dis_and_express.drop_duplicates(subset=['x', 'y'], keep='first')
        for index in dis_and_express.index:
            if (dis_and_express.at[index, 'x'], dis_and_express.at[index, 'y']) in self.inSpots:
                dis_and_express.drop(index=index)
        dis_and_express.reset_index(drop=True)
        return dis_and_express

    def MannWhitney(self,
                    mask: np.ndarray,
                    location: str = 'out',
                    alternative: str = 'two-sided',
                    ran: Optional[float] = None,
                    cut: Optional[float] = 0,
                    alpha: float = 0.05,
                    drop_zero: bool = False
                    ) -> pd.DataFrame:

        from scipy import stats
        from statsmodels.stats.multitest import fdrcorrection

        self.__structureCheck()
        df_dist_express = self.get_dist_and_express(mask, location=location)
        distance = df_dist_express['distance'].tolist()
        near = [True if i > cut else False for i in distance]
        far = [True if i <= cut else False for i in distance]

        df_dist_express.drop(labels=['x', 'y', 'distance', 'distance_abs'], axis=1)
        df_Wilcoxon_Mann_Whitney = pd.DataFrame(columns=['gene', 'stat', 'P value',
                                                         'effective spot near', 'effective spot far'])

        for gene in self.gene_name:
            list_gene_exp = df_dist_express[gene].values
            if ran is not None:
                if max(list_gene_exp) - min(list_gene_exp) < ran:
                    continue
            list_exp_near = list_gene_exp[np.array(near)]
            list_exp_far = list_gene_exp[np.array(far)]
            if drop_zero:
                list_exp_near = list(filter(lambda x: x != 0, list_exp_near))
                list_exp_far = list(filter(lambda x: x != 0, list_exp_far))
                effective_spot_near = len(list_exp_near)
                effective_spot_far = len(list_exp_far)
                if len(list_exp_near) == 0:
                    list_exp_near.append(0)
                if len(list_exp_far) == 0:
                    list_exp_far.append(0)
            else:
                effective_spot_near = len(list_exp_near)
                effective_spot_far = len(list_exp_far)
            stat, pvalue = stats.mannwhitneyu(list_exp_near, list_exp_far, alternative=alternative)
            df_Wilcoxon_Mann_Whitney.loc[len(df_Wilcoxon_Mann_Whitney.index)] = [gene, stat, pvalue,
                                                                                 effective_spot_near,
                                                                                 effective_spot_far]
        rej, p_adjust = fdrcorrection(df_Wilcoxon_Mann_Whitney['P value'], alpha=alpha)
        df_Wilcoxon_Mann_Whitney['P fdr'] = p_adjust
        df_Wilcoxon_Mann_Whitney['P rej'] = rej

        return df_Wilcoxon_Mann_Whitney

    def spearman(self,
                 mask: np.ndarray,
                 num: Optional[float] = None,
                 ran: Optional[float] = None,
                 drop_zero: bool = False
                 ) -> pd.DataFrame:

        from scipy import stats
        from statsmodels.stats.multitest import fdrcorrection

        gap = self.radius / num

        self.__structureCheck()
        df_dist_express = self.get_dist_and_express(mask)

        distance = df_dist_express['distance'].tolist()

        df_dist_express.drop(labels=['x', 'y', 'distance', 'distance_abs'], axis=1)
        df_Spearman = pd.DataFrame(columns=['gene', 'coef', 'P value'])

        dis_level = gap
        level = 1
        for index in range(len(distance)):
            if distance[index] < dis_level:
                distance[index] = level
            else:
                while (distance[index] > dis_level):
                    dis_level += gap
                    level += 1
                distance[index] = level

        for gene in self.gene_name:
            list_gene_exp = df_dist_express[gene].tolist()
            if ran is not None:
                if max(list_gene_exp) - min(list_gene_exp) < ran:
                    continue
            distance_gene = copy.deepcopy(distance)
            if drop_zero:
                length = len(distance)
                drop = 0
                for index in range(length):
                    if list_gene_exp[index - drop] == 0:
                        list_gene_exp.pop(index - drop)
                        distance_gene.pop(index - drop)
                        drop += 1

            coef, pvalue = stats.spearmanr(distance_gene, list_gene_exp)
            df_Spearman.loc[len(df_Spearman.index)] = [gene, coef, pvalue]
        rej, p_adjust = fdrcorrection(df_Spearman['P value'], alpha=0.05)
        df_Spearman['P fdr'] = p_adjust
        df_Spearman['P rej'] = rej
        _add_info_from_sample(self.adata, sample_id=None, keys='spearman', add=df_Spearman)

        return df_Spearman


#####################################################

class GeneCluster(object):
    def __init__(self,
                 dic_crd: dict,
                 method: str,
                 params: pd.DataFrame,
                 norm: bool = True,
                 pvalue: Optional[float] = None,
                 fdr: bool = False,
                 range_min: Optional[float] = None,
                 correlation: Optional[float] = None,
                 ):
        """
        initialization
        """
        self.dic_crd = dic_crd
        self.method = method
        self.df_params = params
        self.norm = norm
        self.pvalue = pvalue
        self.fdr = fdr
        self.range_min = range_min
        self.correlation = correlation
        self.gene_express = pd.DataFrame(dic_crd)
        self.gene_express = self.gene_express.drop_duplicates(subset=['Xest'])
        num_orginal = self.df_params.shape[0]
        self.__select_gene()
        num_filter = self.df_params.shape[0]
        logg.info(f'The original {num_orginal} gene curves, after filtering the remaining {num_filter}')

    def clusterTheGeneTendency(self, k: int, sdbw: bool = True, num=50):
        """
        Determining Regression types
        """

        if self.method == 'lowess':
            self.__cluster_loess(k, sdbw)
        elif self.method == 'poly':
            self.__cluster_Poly(num, k, sdbw)

    def __select_gene(self):
        """
        Filter genes based on user-specified metrics
        """
        if self.pvalue is not None:
            from statsmodels.stats.multitest import fdrcorrection
            if self.fdr:
                rej, p_adjust = fdrcorrection(self.df_params['p_value'], alpha=self.pvalue)
                self.df_params['p_adjust'] = p_adjust
                self.df_params['p_rej'] = rej
                self.df_params = self.df_params[self.df_params['p_rej'] == True]
            else:
                self.df_params = self.df_params[self.df_params['p_value'] < self.pvalue]

        if self.range_min is not None:
            self.df_params = self.df_params[abs(self.df_params['range']) > self.range_min]

        if self.correlation is not None:
            self.df_params = self.df_params[abs(self.df_params['correlation']) > self.correlation]
        assert self.df_params.shape[0] >= 2, 'The number of cluster must >= 2'

    def __cluster_Poly(self, num, k, sdbw):
        """
        Cluster polynomial regression curves
        """

        from .utils import _best_k
        list_gene = []
        x = self.gene_express['Xest']

        for gene_name in self.df_params.index:
            y = self.gene_express[gene_name]
            params = self.df_params.loc[gene_name, 'param']
            xgrid = np.linspace(x.min(), x.max(), num)
            ygrid = []
            for x0 in xgrid:
                ygrid.append(sum([param * x0 ** i for i, param in enumerate(params)]))
            if self.norm:
                ygrid = (np.array(ygrid) - y.min()) / (y.max() - y.min())
            list_gene.append(ygrid)

        array_gene = np.array(list_gene)

        res, km = _best_k(array_gene, k, sdbw)

        self.df_params['cluster'] = res
        self.k = km

    def __cluster_loess(self, k, sdbw):
        """
        Cluster loess regression curves
        """

        from .utils import _best_k
        list_gene = []

        for gene_name in self.df_params.index:
            y = self.gene_express[gene_name]
            if self.norm:
                ygrid = (np.array(y) - y.min()) / (y.max() - y.min())
            else:
                ygrid = y
            list_gene.append(ygrid)

        array_gene = np.array(list_gene)

        res, km = _best_k(array_gene, k, sdbw)
        self.df_params['cluster'] = res
        self.k = km


def adata_from_mask(
        adata: anndata.AnnData,
        mask: np.ndarray,
        location: Literal['all', 'in', 'out'] = 'in',
        spatial_in_obsm: str = 'spatial',
        radius: float = None,
) -> anndata.AnnData:
    """
    Generate a new adata object through the selected mask region.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    mask : numpy.ndarray
        A binarized image data of ROI.
    location : str, optional
        'in': The selected spots inside the contours.
        'out': The selected spots outside the contours.
        'all': both of 'in' and 'out'.
    spatial_in_obsm : str, optional
        The key of spatial coordinates in adata.obsm
    radius : float
        The range of the point whose included in the calculation.

    Returns
    -------
    anndata.Anndata object
    """
    adata = _check_adata_type(adata, spatial_in_obsm, inplace=True)

    new_tendency = SpatialTendency(
        adata=adata,
        radius=None,
        gene_name='None',
        scale=1.0,
        spatial_in_obsm=spatial_in_obsm,
    )

    adata = new_tendency.get_new_adata(
        mask=mask,
        location=location,
        radius=radius,
    )

    return adata


def wilcoxon_test(
        adata: anndata.AnnData,
        mask: np.ndarray,
        radius: float,
        gene_name: Union[str, list, None] = None,
        clusters: Union[str, int, list] = 'all',
        cluster_label: Optional[str] = None,
        location: Literal['all', 'in', 'out'] = 'all',
        scale: Union[float, str] = 'hires',
        spatial_in_obsm: str = 'spatial',
        alternative: str = 'two-sided',
        ran: Optional[float] = None,
        cut: Optional[float] = 0,
        alpha: float = 0.05,
        drop_zero: bool = False,
        inplace: bool = True,
) -> pd.DataFrame:
    """
    Wilcoxon tests were performed using the distance from each spot(cell) to the mask boundary and the
    expression of a gene(protein) at each spot.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    mask : numpy.ndarray
        A binarized image data of ROI.
    radius : float
        Maximum distance considered, negative distance by absolute value.
    gene_name : Union[list, str, None], optional
        The gene names for the regression model need to be calculated.
    clusters : Union[str, int, list], optional
        The cluster of the spot being counted, the default is all clusters.
    cluster_label : str, optional
        The label of cluster in adata.obs.
    location : Literal['all', 'in', 'out']
        'in': The selected spots inside the contours.
        'out': The selected spots outside the contours.
        'all': both of 'in' and 'out'.
    scale : Union[str, float], optional
        Scale used in subsequent analyses. If it's Visium data it can also be HE image labels (hires or lower).
        Most of the time you don't need to change this
    spatial_in_obsm : str, optional
        The key of spatial coordinates in adata.obsm
    alternative: {‘two-sided’, ‘less’, ‘greater’}, optional
        Defines the alternative hypothesis. Default is 'two-sided'.
        For more details, see scipy.stats.mannwhitneyu()
    ran : float, optional
        Minimum range, genes are ignored when their range is smaller than this value.
    cut : float, optional
        Distance threshold for grouping. The default value is 0, which represents the two groups inside and outside the
        contour.
    alpha : float, optional
        Family-wise error rate of P-fdr. Defaults to 0.05.
    drop_zero : bool, optional
        Whether to remove all spots with 0 expression.
    inplace : bool, optional
        Whether to change the original adata.

    Returns
    -------
    pandas.DataFrame
        The wilcoxon-test result is stored, including the p-value('P value'), the fdr p-value('P fdr'),
        and the label of whether the p-value is significant or not('P rej').
    """
    adata = _check_adata_type(adata, spatial_in_obsm, inplace)

    if gene_name is None:
        gene_name = adata.var_names.tolist()

    new_tendency = SpatialTendency(adata=adata,
                                   gene_name=gene_name,
                                   radius=radius,
                                   scale=scale,
                                   clusters=clusters,
                                   cluster_key=cluster_label,
                                   spatial_in_obsm=spatial_in_obsm,
                                   )

    df_wilcoxon = new_tendency.MannWhitney(mask=mask,
                                           location=location,
                                           alternative=alternative,
                                           ran=ran,
                                           cut=cut,
                                           alpha=alpha,
                                           drop_zero=drop_zero,
                                           )

    return df_wilcoxon


def spearman_correlation(
        adata: anndata.AnnData,
        mask: np.ndarray,
        radius: float,
        gene_name: Union[str, list] = 'all',
        clusters: Union[str, int, list] = 'all',
        cluster_key: Optional[str] = None,
        scale: Union[float, str] = 'hires',
        spatial_in_obsm: str = 'spatial',
        num: Optional[float] = None,
        ran: Optional[float] = None,
        drop_zero: bool = False,
        inplace: bool = True,
) -> pd.DataFrame:
    """
    Spearman test was performed on the expression levels of genes in each class after the samples were classified
    equally by distance

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    mask : numpy.ndarray
        A binarized image data of ROI
    radius : float
        Maximum distance considered, negative distance by absolute value
    gene_name : Union[list, str, None], optional
        The gene names for the regression model need to be calculated.
    clusters : Union[str, int, list], optional
        The cluster of the spot being counted, the default is all clusters.
    cluster_label : str, optional
        The label of cluster in adata.obs.
    scale : Union[str, float], optional
        Scale used in subsequent analyses. If it's Visium data it can also be HE image labels (hires or lower).
        Most of the time you don't need to change this
    spatial_in_obsm : str, optional
        The key of spatial coordinates in adata.obsm
    num : int, optional
        The number of gap
    ran : float, optional
        Minimum range, genes are ignored when their range is smaller than this value.
    drop_zero : bool, optional
        Whether to remove all spots with 0 expression.
    inplace : bool, optional
        Whether to change the original adata.

    Returns
    -------
    pandas.DataFrame
        The spearman result is stored, including the p-value('P value'), the fdr p-value('P fdr'),
        and the label of whether the p-value is significant or not('P rej').
    """
    adata = _check_adata_type(adata, spatial_in_obsm, inplace)

    if gene_name == 'all':
        gene_name = adata.var_names.tolist()

    New_tendency = SpatialTendency(adata=adata,
                                   gene_name=gene_name,
                                   radius=radius,
                                   scale=scale,
                                   clusters=clusters,
                                   cluster_key=cluster_key,
                                   spatial_in_obsm=spatial_in_obsm,
                                   )

    df_spearman = New_tendency.spearman(mask=mask,
                                        num=num,
                                        ran=ran,
                                        drop_zero=drop_zero,
                                        )

    return df_spearman


def ANOVA(
        adata: ad.AnnData,
        score_label: str,
        cluster_label: str = 'cell_type',
):
    """
    Anova of genes or other metrics

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    score_label : str
        Selected genes or other indicators in obs.columns.
    cluster_label : str, optional
        The label of cluster in adata.obs.

    Returns
    -------
    The p-value of ANOVA
    """
    from statsmodels.formula.api import ols
    from statsmodels.stats.anova import anova_lm

    cluster = adata.obs[cluster_label]
    if score_label in adata.obs.columns:
        score = np.array(adata.obs[score_label])
    elif score_label in adata.var_names:
        index = adata.var_names.tolist().index(score_label)
        score = adata.X.toarray()[:, index]
    else:
        raise ValueError

    df = pd.DataFrame({'cluster': cluster, 'score': score})

    fml = 'score' + '~C(' + 'cluster' + ')'

    model = ols(fml, data=df).fit()
    anova_table_1 = anova_lm(model, typ=2).reset_index()
    p1 = anova_table_1.loc[0, 'PR(>F)']
    _add_info_from_sample(adata, sample_id=None, keys='ANOVA', add=p1)

    return p1


def spatial_tendency(
        adata: ad.AnnData,
        mask: np.ndarray,
        radius: int,
        method: Literal['poly', 'loess'] = 'poly',
        gene_name: Union[str, list] = 'all',
        clusters: Union[str, int, list] = 'all',
        cluster_key: Optional[str] = None,
        location: Literal['all', 'in', 'out'] = 'all',
        scale: Union[float, str] = 'hires',
        spatial_in_obsm: str = 'spatial',
        frac: Union[int, float] = None,
        sd: Optional[int] = None,
        drop_zero: bool = False,
        inplace: bool = True,
) -> anndata.AnnData:
    """
    We used two regression methods, Loess regression, and Polynomial regression, to study the variation of the
    expression with the min distance from its location to the pixel of boundary.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    mask : numpy.ndarray
        Binarized image data of ROI.
    radius : float
        The range of the point whose included in the calculation.
    method : Literal['poly', 'loess'], optional
        Polynomial regression(poly) or Loess regression(loess).
    gene_name : Union[list, str, None], optional
        The gene names for the regression model need to be calculated.
    clusters : Union[str, int, list], optional
        The cluster of the spot being counted, the default is all clusters.
    cluster_key : Union[str, int, list], optional
        The key of cluster in adata.obs.
    location : str, optional
        'in': The selected spots inside the contours.
        'out': The selected spots outside the contours.
        'all': both of 'in' and 'out'.
    scale : Union[str, float], optional
        scale used in subsequent analyses. If it's Visium data it can also be HE image labels (hires or lower).
        Most of the time you don't need to change this
    spatial_in_obsm : str, optional
        The key of spatial coordinates in adata.obsm
    frac : Union[int, float], optional
        The highest degree of a polynomial regression or lowess regression smoothness.
    sd : int, optional
        The coefficient of the standard deviation of the tail treatment.
    drop_zero : bool, optional
        Whether to remove all spots with 0 expression
    inplace : bool, optional
        Whether to change the original adata.

    Returns
    -------
    - :attr:`anndata.AnnData.uns` ``['SOAPy']['poly']['dic_crd_poly'] or ['SOAPy']['loess']['dic_crd_loess']`` - Store
        the shape of curves
    - :attr:`anndata.AnnData.uns` ``['SOAPy']['poly']['df_param_poly'] or ['SOAPy']['loess']['df_param_loess']`` - Store
        additional params for curves


    """
    adata = _check_adata_type(adata, spatial_in_obsm, inplace)

    if gene_name == 'all':
        gene_name = adata.var_names.tolist()

    New_tendency = SpatialTendency(adata=adata,
                                   gene_name=gene_name,
                                   radius=radius,
                                   scale=scale,
                                   clusters=clusters,
                                   cluster_key=cluster_key,
                                   spatial_in_obsm=spatial_in_obsm,
                                   )
    if method == 'poly':
        if frac is None:
            frac = 4
        adata = New_tendency.polynomialRegression(mask=mask,
                                                  frac=frac,
                                                  sd=sd,
                                                  location=location,
                                                  drop_zero=drop_zero,
                                                  )
    elif method == 'loess':
        if frac is None:
            frac = 0.6
        adata = New_tendency.loess(mask=mask,
                                   frac_0_1=frac,
                                   sd=sd,
                                   location=location,
                                   drop_zero=drop_zero
                                   )
    else:
        logg.error(f'{method} is not in [\'poly\', \'loess\']', exc_info=True)
        raise ValueError()

    return adata


def gene_cluster(
        adata: ad.AnnData,
        k: int,
        norm: bool = True,
        method: Literal['poly', 'loess'] = 'poly',
        pvalue: Optional[float] = None,
        fdr: bool = False,
        range_min: Optional[float] = None,
        correlation: Optional[float] = None,
        best_k_sdbw: bool = True,
        num_spots: int = 50,
        inplace: bool = True,
) -> anndata.AnnData:
    """
    The regression curves of multiple genes are clustered, and the curves can be screened by adjusting a variety of
    indicators

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    k : int
        The number of cluster
    norm : bool, optional
        If true, data normalization
    method : Literal['poly', 'loess'], optional
        Polynomial regression(poly) and Lowess regression(loess).
    pvalue : float, optional
        Threshold for the p-value in Polynomial regression.
    fdr : bool, optional
        If True, the p-values were corrected for FDR.
    range_min : float, optional
        Genes whose expression range is less than this value are discarded.
    correlation : float, optional
        Threshold for the correlation
    best_k_sdbw : bool, optional
        Automated cluster number screening using sdbw.
    num_spots : int, optional
        The number of sampling points in curves clustering.
    inplace : bool, optional
        Whether to change the original adata.

    Returns
    -------
    - :attr:`anndata.AnnData.uns` ``['SOAPy']['poly']['df_param_poly'] or ['SOAPy']['loess']['df_param_loess']`` - Store
        additional params for curves

    """
    adata = _check_adata_type(adata, 'spatial', inplace)

    if method == 'poly':
        poly_data = _get_info_from_sample(adata, sample_id=None, key='poly')
        dic_crd = poly_data['dic_crd_poly']
        params = poly_data['df_param_poly']
    elif method == 'loess':
        loess_data = _get_info_from_sample(adata, sample_id=None, key='loess')
        dic_crd = loess_data['dic_crd_loess']
        params = loess_data['df_param_loess']
    else:
        logg.error(f'{method} is not in [\'poly\', \'loess\']', exc_info=True)
        raise ValueError()

    New_gc = GeneCluster(dic_crd=dic_crd,
                         method=method,
                         params=params,
                         norm=norm,
                         pvalue=pvalue,
                         fdr=fdr,
                         range_min=range_min,
                         correlation=correlation)
    New_gc.clusterTheGeneTendency(k=k, sdbw=best_k_sdbw, num=num_spots)

    gene_cluster_params = {'gene_cluster': New_gc.df_params, 'k': New_gc.k, 'method': method}
    _add_info_from_sample(adata, sample_id=None, keys='gene_cluster', add=gene_cluster_params)

    return adata



import anndata
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap
from os import PathLike
from types import MappingProxyType
from typing import Optional, Union, Any, Tuple, Sequence, Mapping
import copy
from ._color import _get_palette
from ..utils import _scale, _get_info_from_sample
from matplotlib.axes import Axes

__all__ = [
    'show_moran_scatterplot',
    'show_network',
    'show_voronoi'
]


def show_moran_scatterplot(
        adata: anndata.AnnData,
        palette: Union[Sequence[str], ListedColormap] = None,
        figsize: Optional[Tuple[float, float]] = None,
        dpi: Optional[float] = None,
        title: Optional[str] = None,
        ax: Optional[Axes] = None,
        show: bool = True,
        save: Union[str, PathLike] = None,
        **kwargs
) -> Optional[Axes]:
    """
    Show a Moran scatter plot. Highlight the hotspots of the domain_from_local_moran.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    figsize : Tuple[float, float], optional
        (Width, height) of the figure.
    dpi : float, optional
        The resolution of the figure.
    title : str, optional
        The title of shown figure.
    ax : Axes
        A matplotlib axes object.
    show : bool
        Show the plot, do not return axis.
    save : Union[str, PathLike], optional
        The path where the image is stored.
    palette : Union[Sequence[str], Cycler], optional
        (hotspot_color, other_color).
        Displays the Moran index color for hot and non-hot areas, inputing a list of two RGB colors.
    **kwargs
        Other params of ax.scatter()

    Returns
    -------
        If show==False, return Axes.
    """
    if palette is None:
        palette = ['#dd7e6b', '#bababa']
    if isinstance(palette, ListedColormap):
        palette = palette.colors[0:2]

    if ax is None:
        if figsize is None:
            figsize = (7, 7)

        fig = plt.figure(figsize=figsize, dpi=dpi)
        ax = fig.add_subplot(111)

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)

    ax.set_xlabel('Attribute')
    ax.set_ylabel('Spatial Lag')
    ax.set_title('Moran Scatterplot')

    local_moran = _get_info_from_sample(adata=adata, sample_id=None, key='local_moran')
    local_moran_df = local_moran['LocalMoran']

    for index, spot in local_moran_df.iterrows():
        x = spot['Attribute']
        y = spot['Spatial Lag']
        label = spot['Label']
        if label == 'Hotspot':
            c = palette[0]
        else:
            c = palette[1]
        ax.scatter(x, y,
                   label=label,
                   c=c,
                   **kwargs
                   )

    handles, labels = ax.get_legend_handles_labels()
    unique = [(h, l) for i, (h, l) in enumerate(zip(handles, labels)) if l not in labels[:i]]
    ax.legend(*zip(*unique))
    attribute_filter = local_moran['attribute_filter']
    spatial_lag_filter = local_moran['spatial_lag_filter']

    ax.vlines(attribute_filter, min(local_moran_df['Spatial Lag']), max(local_moran_df['Spatial Lag']), linestyle='--')
    ax.hlines(spatial_lag_filter, min(local_moran_df['Attribute']), max(local_moran_df['Attribute']), linestyle='--')

    if title is None:
        title = 'Moran Scatterplot'
    plt.title(title, fontsize=20)

    if save:
        plt.savefig(save)

    if show:
        plt.show()
    else:
        return ax


def show_voronoi(
        adata: anndata.AnnData,
        cluster_key: str,
        sample_key: Optional[str] = 'sample',
        sample_id: Union[int, str] = None,
        palette: Union[Sequence[str], ListedColormap, dict] = None,
        title: Optional[str] = None,
        spatial_in_obsm: str = 'spatial',
        figsize: Optional[Tuple[float, float]] = None,
        dpi: Optional[float] = None,
        ax: Optional[Axes] = None,
        show: bool = True,
        save: Union[str, PathLike] = None,
        legend_kwargs: dict = {},
        **kwargs: Any
) -> Optional[Axes]:
    """
    Draw a Voronoi plot of the cell distribution.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    figsize : Tuple[float, float], optional
        (Width, height) of the figure.
    dpi : float, optional
        The resolution of the figure.
    title : str, optional
        The title of shown figure.
    ax : Axes
        A matplotlib axes object.
    show : bool
        Show the plot, do not return axis.
    save : Union[str, PathLike], optional
        The path where the image is stored.
    sample_key : str, optional
        Batch's key in adata.obs.
    sample_id : Union[int, str], optional
        The sample number which to be shown.
    cluster_key : str
        Keys for annotations in adata.obs.
    palette : Union[Sequence[str], Cycler, dict]
        Colors to use for plotting annotation groups.
    spatial_in_obsm : str
        Keyword of coordinate information in obsm.
    legend_kwargs : dict, optional
        Other params of plt.legend().
    kwargs : Any, optional
        Other params of scipy.spatial.voronoi_plot_2d().

    Returns
    -------
        If show==False, return Axes.
    """
    import matplotlib.pyplot as plt
    from scipy.spatial import Voronoi, voronoi_plot_2d
    from collections import OrderedDict

    if sample_id is not None:
        adata = adata[adata.obs[sample_key] == sample_id, :].copy()
    else:
        adata = copy.deepcopy(adata)

    if title is None:
        title = str(cluster_key)

    points = adata.obsm[spatial_in_obsm]
    cluster_list = adata.obs[cluster_key]

    cluster_unique = cluster_list.unique()

    if not isinstance(palette, dict):
        palette = _get_palette(list(cluster_unique), sort_order=True, palette=palette)
    else:
        palette = palette

    points_color = [palette[index] for index in cluster_list]

    vor = Voronoi(points)

    if figsize is None:
        row = np.max(points[:, 1])
        col = np.max(points[:, 0])
        figsize = (8, 8 * row / col)

    if ax is None:
        fig = plt.figure(figsize=figsize, dpi=dpi)
        ax = fig.add_axes([0.1, 0.2, 0.6, 0.6])

    voronoi_plot_2d(vor,
                    show_vertices=False,
                    point_size=0,
                    ax=ax,
                    **kwargs)

    for r in range(len(vor.point_region)):
        region = vor.regions[vor.point_region[r]]
        if not -1 in region:
            polygon = [vor.vertices[i] for i in region]
            plt.fill(*zip(*polygon), color=points_color[r], label=cluster_list[r])
    handles, labels = plt.gca().get_legend_handles_labels()
    by_label = OrderedDict(zip(labels, handles))
    new_label = {}
    for key in sorted(by_label):
        new_label[key] = by_label[key]

    if 'bbox_to_anchor' not in legend_kwargs.keys():
        legend_kwargs['bbox_to_anchor'] = (1.02, 0.8)
    if 'loc' not in legend_kwargs.keys():
        legend_kwargs['loc'] = 'upper left'
    if 'title' not in legend_kwargs.keys():
        legend_kwargs['title'] = cluster_key

    ax.legend(new_label.values(), new_label.keys(), **legend_kwargs)
    ax.set_xticks([])
    ax.set_yticks([])
    plt.title(title, fontsize=20)
    plt.tight_layout()

    if save:
        plt.savefig(save)

    if show:
        plt.show()
    else:
        return ax


def show_network(
        adata: anndata.AnnData,
        sample_key: Optional[str] = 'sample',
        sample_id: Union[int, str] = None,
        cluster_key: str = 'clusters',
        edge_color: str = 'b',
        palette: Union[Sequence[str], ListedColormap, dict] = None,
        title: Optional[str] = 'Spatial network',
        scale: Union[str, float] = 'hires',
        spatial_in_obsm: str = 'spatial',
        figsize: Optional[Tuple[float, float]] = None,
        dpi: int = 100,
        ax: Optional[Axes] = None,
        spot_size: float = 5,
        show: bool = True,
        save: Union[str, PathLike] = None,
        legend_kwargs: dict = {},
        **kwargs: Any
) -> Optional[Axes]:
    """
    Show the spatial network.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    figsize : Tuple[float, float], optional
        (Width, height) of the figure.
    dpi : float, optional
        The resolution of the figure.
    title : str, optional
        The title of shown figure.
    ax : Axes
        A matplotlib axes object.
    show : bool
        Show the plot, do not return axis.
    save : Union[str, PathLike], optional
        The path where the image is stored.
    sample_key : str, optional
        Batch's key in adata.obs.
    sample_id : Union[int, str], optional
        The sample number which to be shown.
    cluster_key : str
        Keys for annotations in adata.obs. It can not be None.
    edge_color : str
        The color of edges in the network.
    palette : Union[Sequence[str], Cycler, dict], optional
        Colors to use for plotting annotation groups.
    scale : Union[str, float]
        The scaling factor for distance scaling. If it's Visium data it can also be HE image labels (hires or lower).
        Most of the time you don't need to change this.
    spatial_in_obsm : str
        Keyword of coordinate information in obsm.
    spot_size : float
        The size of the spot
    legend_kwargs : dict, optional
        Other params of plt.legend().
    kwargs : Any, optional
        Other params of scipy.spatial.voronoi_plot_2d().

    Returns
    -------
        If show==False, return Axes.
    """
    edges = _get_info_from_sample(adata=adata, sample_id=sample_id, key='edges')
    if sample_id is not None:
        bdata = adata[adata.obs[sample_key] == sample_id, :].copy()
    else:
        bdata = copy.deepcopy(adata)
    obs = bdata.obs
    if type(scale) != float:
        scale = _scale(bdata, None, scale)

    df_pixel = bdata.obsm[spatial_in_obsm]
    df_pixel = pd.DataFrame(df_pixel, index=bdata.obs_names) * scale
    obs['x'] = df_pixel[0].tolist()
    obs['y'] = df_pixel[1].tolist()

    if figsize is None:
        row = max(obs['y'])
        col = max(obs['x'])
        figsize = (8, 8 * row / col)

    if ax is None:
        fig = plt.figure(figsize=figsize,
                         dpi=dpi,
                         )
        ax = fig.add_axes([0.1, 0.2, 0.6, 0.6])
    ax.xaxis.set_ticks_position('top')
    ax.invert_yaxis()
    for index, row in edges.iterrows():
        index_i = obs.iloc[int(getattr(row, 'point_1')), :]
        index_j = obs.iloc[int(getattr(row, 'point_2')), :]
        ax.plot([index_i['x'], index_j['x']],
                [index_i['y'], index_j['y']],
                c=edge_color,
                zorder=2,
                **kwargs
                )

    cluster_unique = obs[cluster_key].unique()
    cluster_unique = sorted(cluster_unique)

    if not isinstance(palette, dict):
        palette = _get_palette(list(cluster_unique), sort_order=True, palette=palette)
    else:
        palette = palette

    for index, clu in enumerate(cluster_unique):
        sub_obs = obs[obs[cluster_key] == clu]
        ax.scatter(sub_obs['x'],
                   sub_obs['y'],
                   label=clu,
                   c=palette[clu],
                   s=spot_size,
                   zorder=3)

    if 'bbox_to_anchor' not in legend_kwargs.keys():
        legend_kwargs['bbox_to_anchor'] = (1.02, 0.8)
    if 'loc' not in legend_kwargs.keys():
        legend_kwargs['loc'] = 'upper left'
    if 'title' not in legend_kwargs.keys():
        legend_kwargs['title'] = cluster_key

    ax.legend(**legend_kwargs)

    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_title(label=title, fontsize=20)
    plt.tight_layout()

    if save:
        plt.savefig(save)

    if show:
        plt.show()
    else:
        return ax


"""
Spatial Omics includes barcode-based spatial transcriptomics, image-based spatial transcriptomics ,spatial proteomics and
spatial metabolomics. Barcode-based data is easily converted to Anndata format, but image-based data need Pre-processing to
convert to Anndata. We developed a model named all2adata to convert all spatial omics data to Anndata format.
"""

import os
import pandas as pd
import numpy as np
import anndata
import scanpy as sc
import cv2 as cv
import matplotlib.pyplot as plt
from scipy import sparse
from os import PathLike
from typing import Optional, Union, Literal
from pathlib import Path
import scipy.io as sio

__all__ = ["read_csv2adata", "read_mult_image2adata", "read_visium2adata", "read_dsp2adata", "st_Seurat2Anndata"]

###############################Spatial Omics to Anndata##############################


def read_csv2adata(
        express: Union[PathLike, str, pd.DataFrame],
        meta: Union[PathLike, str, pd.DataFrame] = None,
        spatial: Union[np.ndarray, list] = None,
        express_kwargs: dict = {},
        meta_kwargs: dict = {},
) -> anndata.AnnData:
    """
    To read Barcode_based spatial omics data.
    Slide seqV2, Stereo seq and other table data

    Parameters
    ----------
    express : Union[PathLike, str, pd.DataFrame]
        Count matrix file or its path.
    location : Union[PathLike, str, pd.DataFrame]
        Coordinates matrix file or its path.
    scale : float
        The scale of the coordinates, usually for visual matching with different resolutions of HE or other images,
        is not scaled by default.
    express_kwargs : dict
        Other params of pd.read_csv(express)
    location_kwargs : dict
        Other params of pd.read_csv(location)

    Returns
    -------
        anndata.Anndata object
    """

    if type(express) is not pd.DataFrame:
        express = pd.read_csv(express, **express_kwargs)

    if type(meta) is not pd.DataFrame and meta is not None:
        meta = pd.read_csv(meta, **meta_kwargs)

    adata = sc.AnnData(
        express.values,
        obs=meta,
        var=pd.DataFrame(index=express.columns),
        obsm={},
    )
    # sc.pp.calculate_qc_metrics(adata, inplace=True)

    if isinstance(spatial, np.ndarray):
        adata.obsm['spatial'] = spatial
    elif isinstance(spatial, list):
        if len(spatial) != 2:
            raise ValueError('The data dimension must be equal to 2')
        i, j = spatial
        assert (i in list(meta.columns) and j in list(meta.columns)), 'The coordinate information should be included in the meta'
        adata.obsm['spatial'] = meta.loc[:, [i, j]].values

    adata.var_names_make_unique()
    adata.X = sparse.csr_matrix(adata.X)

    return adata


def read_mult_image2adata(
        image: Union[np.ndarray, str, PathLike],
        mask: Union[np.ndarray, str, PathLike],
        channel_names: list = None,
        remove_channels: list = None,
        max_quantile: float = 0.98,
        scale: float = 1.0,
) -> anndata.AnnData:
    """
    To read Image_based spatial omics data.
    Slide seqV2, Stereo seq and other image data

    Parameters
    ----------
    image : Union[np.ndarray, str, PathLike]
        multiplexed image (.tiff file); with dimension (z,x,y), where z represents channel number, x and y represent length and width of sample respectively
        you can use np.swapaxes() for dimension. E.g. convert (x,y,z) to (z,x,y):
        img_new=np.swapaxes(np.swapaxes(img,0,2),1,2) # (x,y,z)→(z,y,x)→(z,x,y).
    mask : Union[np.ndarray, str, PathLike]
        image of mask, whose size should be equal to img.
    channel_names : list
        list of channel names (e.g. protein names), whose length should be equal to channel number of img.
    remove_channels : list
        List of the index of channels to be removed.
    max_quantile : float
        highest quantile clipped as highest value.
    scale ; float
        The scale of the coordinates, usually for visual matching with different resolutions of HE or other images,
        is not scaled by default.

    Returns
    -------
        anndata.Anndata object
    """

    import tifffile

    if isinstance(image, np.ndarray):
        img = image.astype(np.uint8)
    else:
        img = tifffile.imread(image)

    if type(mask) is np.ndarray:
        mask = mask.astype(np.uint8)
    else:
        mask = plt.imread(mask, 0)

    mask[mask != 0] = 1
    retval, mask_new, stats, centroids = cv.connectedComponentsWithStats(mask, connectivity=8)
    mask = mask_new
    exp_mat = np.zeros((mask.max(), img.shape[0]))

    for g in range(img.shape[0]):
        img_g = img[g, :, :]
        if img_g.sum() != 0:
            img_g[img_g >= np.quantile(img_g[img_g != 0], max_quantile)] = np.quantile(img_g[img_g != 0],
                                                                                       max_quantile)
        for i in range(1, mask.max() + 1):
            l = mask == i
            exp_mat[i - 1, g] = np.mean(img_g[l])

    if remove_channels:
        for g in remove_channels:
            exp_mat[:, g] = 0

    adata = anndata.AnnData(exp_mat)

    if channel_names is None:
        channel_names = [index for index in img.shape[0]]

    if channel_names:
        adata.var_names = channel_names
    adata.obs_names = range(1, len(adata.obs_names) + 1)

    adata.uns['img'] = img
    adata.uns['mask'] = mask
    adata.uns['var_for_analysis'] = [channel_names[i] for i in range(len(channel_names)) if
                                     remove_channels.count(i) == 0]

    adata.obsm['spatial'] = centroids[1:, :]

    adata.var_names_make_unique()

    adata.uns["spatial"] = {}
    adata.uns["spatial"]['scale'] = scale
    adata.uns['SOAPy'] = {}

    return adata


def read_visium2adata(
        path: Union[str, PathLike],
        count_file: str = "filtered_feature_bc_matrix.h5",
        filtered_feature_bc_matrix='filtered_feature_bc_matrix',
) -> anndata.AnnData:
    """
    Read spatial transcriptomics data of 10X Visium.
        If an h5 file exists, the read_visium() method in scanpy is called by default,
        otherwise it is read through the filtered_feature_bc_matrix file.

    filtered_feature_bc_matrix file need include:
        features.tsv.gz;
        barcodes.tsv.gz;
        matrix.mtx.gz.

    Parameters
    ----------
    path :
        Path of Visium root directory folder.
    count_file :
        Which file in the passed directory to use as the count file.
        Typically either filtered_feature_bc_matrix.h5 or raw_feature_bc_matrix.h5.
    filtered_feature_bc_matrix :
        The folder containing the counts information in the root directory.


    Returns
    -------
        anndata.Anndata object
    """
    # sc.read_visium()
    path = Path(path)
    if os.path.exists(path / count_file):
        adata = sc.read_visium(path, count_file=count_file)
    else:
        genes = pd.read_csv(path / filtered_feature_bc_matrix / 'features.tsv.gz', header=None, sep='\t')
        barcodes = pd.read_csv(path / filtered_feature_bc_matrix / 'barcodes.tsv.gz', header=None, sep='\t')
        mtx = sio.mmread(path / filtered_feature_bc_matrix / 'matrix.mtx.gz').T

        genes = pd.DataFrame(index=genes[0].tolist())
        barcodes = pd.DataFrame(index=barcodes[0].tolist())

        crd = pd.read_csv(path / 'spatial/tissue_positions_list.csv', header=None, index_col=0)
        crd = crd.iloc[:, [3, 4]]
        index = np.isin(crd.index.tolist(), barcodes.index.tolist())
        crd = crd.loc[index, :]

        spatial = crd.values
        adata = anndata.AnnData(mtx, obs=barcodes, var=genes, obsm={})

        adata.obsm['spatial'] = spatial

    return adata


def read_dsp2adata(
        information_file: Union[str, PathLike],
        slide_name: Union[str, list] = None,
        xml_file: Optional[dict] = None,
        polygon_key: str = 'Polygon',
        point_attribute: str = 'Points',
) -> anndata.AnnData:
    """
    Read spatial transcriptomics data of NanoString GeoMx DSP.

    The Export3_BiologicalProbeQC.xlsx or Export4_NormalizationQ3.xlsx file in DSP is used to build adata data.
    If the information of each AOI(ROI) sampling point needs to be placed,
    the dictionary is used to pass in the xml file of each slide.

    Parameters
    ----------
    information_file : Union[str, PathLike]
        The path of Export3_BiologicalProbeQC.xlsx or Export4_NormalizationQ3.xlsx file.
    slide_name : Union[str, list], optional
        The name or names list of used slides. By default, all slides are read.
    xml_file : dict, optional
        If sampling points inside each AOI are needed, pass in the xml file using the dictionary.
        e.g.: {'mu_dev_E13_011': './mu_dev_E13_011.ome.xml'}
    polygon_key : str
        Polygon Tag in .xml file
    point_attribute : str
        Point attribute in polygon Tag.

    Returns
    -------
        anndata.Anndata object
    """
    import xml.dom.minidom

    if isinstance(slide_name, str):
        slide_name = [slide_name]

    roi_inf = pd.read_excel(
        information_file,
        sheet_name='SegmentProperties',
        index_col='SegmentDisplayName',
        header=0,
    )
    if slide_name is None:
        obs = roi_inf
    else:
        obs = roi_inf.loc[[True if i in slide_name else False for i in roi_inf['ScanLabel'].tolist()], :]

    roi_express = pd.read_excel(
        information_file,
        sheet_name='TargetCountMatrix',
        index_col=0,
        header=0,
    ).T

    express = roi_express.loc[obs.index, :]
    var = pd.DataFrame(index=express.columns)

    adata = anndata.AnnData(express.values, obs=obs, var=var, uns={}, obsm={})
    adata.obsm['spatial'] = obs.loc[:, ['ROICoordinateX', 'ROICoordinateY']]

    if xml_file is not None:
        slide_list = list(obs['SlideName'].unique())
        points_roi = pd.DataFrame(columns=['slide', 'roi', 'x', 'y'])

        for slide, path_xml in xml_file.items():
            if slide not in slide_list:
                raise ValueError

            dom = xml.dom.minidom.parse(path_xml)
            root = dom.documentElement
            Polygons = root.getElementsByTagName(polygon_key)
            roi_list = obs[obs['SlideName'] == slide]['ROILabel'].astype(int).tolist()

            for index, Polygon in enumerate(Polygons):
                if (index + 1) not in roi_list:
                    continue
                points = Polygon.getAttribute(point_attribute)
                points = points.split()
                for point in points:
                    x, y = point.split(',')
                    x = float(x)
                    y = float(y)
                    points_roi.loc[len(points_roi.index)] = [slide, index + 1, x, y]

        adata.uns['point'] = points_roi

    return adata

# based rpy2
# Python library preparation
from typing import List
from os import PathLike


# function to transform 10X Visium data
def st_Seurat2Anndata(
        seurat_file: PathLike,
        exp_mat_slot: List = ['RNA', 'counts'],
        res_type: str = 'lowres',
) -> anndata.AnnData:
    """
    Transform the Seurat .rds spatial omics data into Anndata.

    Parameters
    ----------
    seurat_file : PathLike
        Path of seurat_obj.rds file.
    exp_mat_slot : list
        list with assay and slot name of expression matrix in seurat object to export to adata.X.
    res_type : str
        resolution type of image in Seurat object (default: 'lowres' in Seurat).
    Returns
    -------
        anndata.Anndata object

    """

    # R console preparation
    from rpy2.robjects import r
    import rpy2.robjects as robjects
    from rpy2.robjects.packages import importr
    from rpy2.robjects import pandas2ri

    # assert r_home!=None 'R Console is not assigned'

    # Automatically transfer R-data.frame to Python-DataFrame
    pandas2ri.activate()

    # R fucntion preparation
    readRDS = robjects.r['readRDS']
    to_array = robjects.r['as.array']
    to_data_frame = robjects.r['as.data.frame']
    transpose = robjects.r['t']

    # required R package loading

    Seurat = importr('Seurat')
    SeuratObject = importr('SeuratObject')

    # get simple data as single cell
    GetMetaData = r("""
    function(seurat_obj){
    library(Seurat)
    metadata=seurat_obj@meta.data
    metadata$Idents=Idents(seurat_obj)
    return(metadata)
    }
    """)

    GetMetaFeature = r("""
    function(seurat_obj,assay='RNA'){
    library(Seurat)
    library(dplyr)
    gene_feature=GetAssay(seurat_obj,assay=assay)@meta.features
    gene_feature$'genome'=1
    return(gene_feature)
    }
    """)

    # GetAssayData = r("""
    # function(seurat_obj,assay,slot){
    # library(Seurat)
    # exp_mat=GetAssayData(seurat_obj,assay=assay,slot=slot)
    # return(exp_mat %>% as.data.frame())
    # }
    # """)

    # get spatial data specific specific for ST
    GetSpatial = r("""
    function(seurat_obj)
    return(seurat_obj@images)
    """)

    GetCoordinates_Visium = r("""
    function(image_obj){
    library(Seurat)
    coord_mat=image_obj@coordinates
    return(coord_mat[,c('imagerow','imagecol')])
    }
    """)

    GetCoordinates_SlideSeq = r("""
    function(image_obj){
    library(Seurat)
    coord_mat=image_obj@coordinates
    return(coord_mat[,c('x','y')])
    }
    """)

    GetCellCoordinates = r("""
    function(image_obj){
    library(Seurat)
    return(GetTissueCoordinates(image_obj[["centroids"]])[,c('x','y')])
    }
    """)

    GetSegmentationCoordinates = r("""
    function(image_obj){
    library(Seurat)
    return(GetTissueCoordinates(image_obj[["segmentation"]]))
    }
    """)

    GetMoleculeCoordinates = r("""
    function(image_obj,assay,slot){
    library(Seurat)
    library(dplyr)
    x=rownames(GetAssayData(seurat_obj,assay=assay,slot=slot))
    return(GetTissueCoordinates(image_obj[["molecules"]]) %>% filter(molecule %in% x))
    }
    """)

    GetArrayCoordinates = r("""
    function(image_obj){
    library(Seurat)
    coord_mat=image_obj@coordinates[,c('col','row')]
    colnames(coord_mat)=c('array_row','array_col')
    return(coord_mat)
    }
    """)

    GetImage = r("""
    function(image_obj){
    library(Seurat)
    return(image_obj@image)
    }
    """)

    GetScaleFactor = r("""
    function(image_obj){
    library(Seurat)
    return(image_obj@scale.factors)
    }
    """)

    GetImageType = r("""
    function(image_obj){
    library(Seurat)
    return(class(image_obj))
    }
    """)

    import pandas as pd
    import numpy as np

    seurat_obj = readRDS(seurat_file)

    assay = exp_mat_slot[0]
    slot = exp_mat_slot[1]
    count_mat = SeuratObject.GetAssayData(seurat_obj, assay=assay, slot=slot)

    count_mat = to_array(count_mat)
    count_mat = count_mat.transpose()

    # get gene names
    metafeature = robjects.conversion.rpy2py(GetMetaFeature(seurat_obj, assay))

    # metadata
    metadata = robjects.conversion.rpy2py(GetMetaData(seurat_obj))

    # create anndata
    adata = anndata.AnnData(X=count_mat, var=metafeature, obs=metadata)
    coord_mat_ls = []

    images = GetSpatial(seurat_obj)
    ImageType = GetImageType(images[0])[0]

    # 10X Visium
    if ImageType == 'VisiumV1':
        array_coord_mat_ls = []
        spatial_dir = {}
        for i in range(len(images)):
            image_obj = images[i]
            # sample name
            sample_name = images.names[i]
            # coordinate matrix
            coord_mat = GetCoordinates_Visium(image_obj)
            coord_mat = np.array(coord_mat)
            coord_mat = coord_mat.T
            coord_mat_ls.append(coord_mat)
            array_coord_mat = robjects.conversion.rpy2py(GetArrayCoordinates(image_obj))
            array_coord_mat_ls.append(array_coord_mat)
            # image array
            image_array = SeuratObject.GetImage(image_obj, mode='raw')
            # scale factor
            scalefactors = GetScaleFactor(image_obj)
            scalefactors = {scalefactors.names[i]: scalefactors[i][0] for i in range(len(scalefactors))}
            scalefactors = {'fiducial_diameter_fullres': scalefactors['fiducial'],
                            'tissue_hires_scalef': scalefactors['hires'],
                            'tissue_lowres_scalef': scalefactors['lowres']}
            scalefactors['spot_diameter_fullres'] = 0.6196 * scalefactors['fiducial_diameter_fullres']
            spatial_dir[sample_name] = {'images': {res_type: image_array},
                                        'scalefactors': scalefactors}
        # add spatial data
        adata.uns['spatial'] = spatial_dir
        if len(array_coord_mat_ls) == 1:
            array_coord_mat = array_coord_mat_ls[0]
        else:
            array_coord_mat = pd.concat(array_coord_mat_ls, axis=0)
        adata.obs = pd.concat((adata.obs, array_coord_mat), axis=1)

    # Slide-Seq
    if ImageType == 'SlideSeq':
        for i in range(len(images)):
            image_obj = images[i]
            # sample name
            sample_name = images.names[i]
            # coordinate matrix
            coord_mat = GetCoordinates_SlideSeq(image_obj)
            coord_mat = np.array(coord_mat)
            coord_mat = coord_mat.T
            coord_mat_ls.append(coord_mat)

    # STARmap
    # if ImageType=='STARmap':

    # MERSCOPe, Xenium, Nanostring CosMx, CODEX, etc.
    if ImageType == 'FOV':
        for i in range(len(images)):
            image_obj = images[i]
            # sample name
            sample_name = images.names[i]
            # coordinate matrix
            coord_mat = GetCellCoordinates(image_obj)
            coord_mat = np.array(coord_mat)
            coord_mat = coord_mat.T
            coord_mat_ls.append(coord_mat)
            # image_data=GetMoleculeCoordinates(image_obj,assay,slot)
            # segmentation_data=GetSegmentationCoordinates=(image_obj)

    # add coordination matrix
    if len(coord_mat_ls) == 1:
        coord_mat = coord_mat_ls[0]
    else:
        coord_mat = np.concatenate(coord_mat_ls, axis=0)
    coord_mat = coord_mat[:, [1, 0]]
    adata.obsm['spatial'] = coord_mat

    return adata


# function to transform scRNA-Seq data
def sc_Seurat2Anndata(
        seurat_file,
        exp_mat_slot=['RNA', 'counts']
):
    # R console preparation
    from rpy2.robjects import r
    import rpy2.robjects as robjects
    from rpy2.robjects.packages import importr
    from rpy2.robjects import pandas2ri
    # assert r_home!=None 'R Console is not assigned'

    # Automatically transfer R-data.frame to Python-DataFrame
    pandas2ri.activate()

    # R fucntion preparation
    readRDS = robjects.r['readRDS']
    to_array = robjects.r['as.array']
    to_data_frame = robjects.r['as.data.frame']
    transpose = robjects.r['t']

    # required R package loading

    Seurat = importr('Seurat')
    SeuratObject = importr('SeuratObject')

    # get simple data as single cell
    GetMetaData = r("""
    function(seurat_obj){
    library(Seurat)
    metadata=seurat_obj@meta.data
    return(metadata)
    }
    """)

    GetMetaFeature = r("""
    function(seurat_obj,assay='RNA'){
    library(Seurat)
    library(dplyr)
    gene_feature=GetAssay(seurat_obj,assay=assay)@meta.features
    gene_feature$'genome'=1
    return(gene_feature)
    }
    """)

    GetAssayData = r("""
    function(seurat_obj,assay,slot){
    library(Seurat)
    exp_mat=GetAssayData(seurat_obj,assay=assay,slot=slot)
    return(exp_mat %>% as.data.frame())
    }
    """)

    seurat_obj = readRDS(seurat_file)

    # raw count matrix
    assay = exp_mat_slot[0]
    slot = exp_mat_slot[1]
    count_mat = SeuratObject.GetAssayData(seurat_obj, assay=assay, slot=slot)
    count_mat = to_array(count_mat)
    count_mat = count_mat.transpose()

    # get gene names
    metafeature = GetMetaFeature(seurat_obj, assay)

    # metadata
    metadata = GetMetaData(seurat_obj)

    # create anndata
    adata = anndata.AnnData(X=count_mat, var=metafeature, obs=metadata)

    return adata


import umap
import anndata
import numpy as np
from scipy.spatial import distance_matrix
from sklearn.manifold import MDS, smacof

from .._utils import graph_alpha
from .._utils import graph_knn
from .._utils import estimate_cutoff_knn
from .._utils import rep_dgi
from .._utils import rep_gae

def spatial_graph(
    adata: anndata.AnnData,
    method="alpha shape", 
    alpha_n_layer=1, 
    cut=None, 
    estimate_cut=True, 
    knn_n_neighbors=10,
    draw = False
):
    pts = adata.obsm['spatial']
    if estimate_cut:
        cut = estimate_cutoff_knn(pts, k=knn_n_neighbors)
    else:
        cut = np.inf
    if method == "alpha shape":
        A = graph_alpha(pts, cut=cut, n_layer=alpha_n_layer, draw=draw)
    elif method == "knn":
        A = graph_knn(pts, cut=cut, k=knn_n_neighbors, draw=draw)
    
    adata.obsp['scanit-graph'] = A

# def spatial_deconvolution(
#     adata: anndata.AnnData,
#     method='laplacian'
# ):


def spatial_representation(
    adata: anndata.AnnData,
    n_h = 32,
    model = 'dgi',
    n_epoch = 1000,
    lr = 0.001,
    print_step = 500,
    torch_seed = None,
    python_seed = None,
    numpy_seed = None,
    device=None,
    data_slot = None,
    n_consensus = 1,
    projection = 'mds',
    n_comps_proj = 15,
    n_nb_proj = 15,
    extra_embeddings = None,
    extra_embedding_weights = None
):
    A = adata.obsp['scanit-graph']
    if not data_slot is None:
        X_processed = np.array( adata.obsm[data_slot] )
    else:
        X_processed = np.array( adata.X )
        
    if model == "dgi":
        if n_consensus == 1 and extra_embeddings is None:
            X_embed = rep_dgi(n_h, X_processed, A, 
                n_epoch=n_epoch, lr=lr, print_step=print_step, 
                torch_seed=torch_seed, python_seed=python_seed,
                numpy_seed=numpy_seed, device=device)
        else:
            X_embeds = []

            np.random.seed(torch_seed)
            torch_seeds = np.random.choice(10000, size=n_consensus, replace=False)
            np.random.seed(python_seed)
            python_seeds = np.random.choice(10000, size=n_consensus, replace=False)
            np.random.seed(numpy_seed)
            numpy_seeds = np.random.choice(10000, size=n_consensus, replace=False)
            
            for i in range(n_consensus):
                X_embed = rep_dgi(n_h, X_processed, A, 
                    n_epoch=n_epoch, lr=lr, print_step=print_step, 
                    torch_seed=torch_seeds[i], python_seed=python_seeds[i],
                    numpy_seed=numpy_seeds[i], device=device)
                X_embeds.append(X_embed)
            if not extra_embeddings is None:
                for extra_embedding in extra_embeddings:
                    X_embeds.append(extra_embedding)
            
            if extra_embedding_weights is None:
                embeds_weights = np.ones(len(X_embeds)) / float(len(X_embeds))
            else:
                embeds_weights = []
                for i in range(n_consensus):
                    embeds_weights.append( (1-np.sum(extra_embedding_weights))/float(n_consensus) )
                embeds_weights.extend(extra_embedding_weights)
                embeds_weights = np.array(embeds_weights, float)
    elif model == "gae":
        X_embed = rep_gae(n_h, X_processed, A, n_epoch=n_epoch)

    if n_consensus > 1 or not extra_embeddings is None:
        n_spot = X_embed.shape[0]
        W_consensus = np.zeros([n_spot, n_spot])
        for i in range(len(X_embeds)):
            W = distance_matrix(X_embeds[i], X_embeds[i])
            W_consensus += W * embeds_weights[i]
        if projection == 'mds':
            # X_embed,_ = smacof(W_consensus, n_components=n_comps_proj, n_jobs=-1)
            model = MDS(n_components=n_comps_proj, dissimilarity='precomputed', n_jobs=-1, random_state=python_seed)
            X_embed = model.fit_transform(W_consensus)
        elif projection == 'umap':
            model = umap.UMAP(n_components=n_comps_proj, metric='precomputed', n_neighbors=n_nb_proj)
            X_embed = model.fit_transform(W_consensus)

    adata.obsm['X_scanit'] = X_embed
    if n_consensus > 1:
        adata.obsp['D_scanit'] = W_consensus


import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from scanpy.plotting import palettes


def transparent_back(img: np.ndarray):
    """
    Set the image's white background to transparent
    Parameters
    ----------
    img: image of numpy.ndarray

    Returns
    -------

    """
    from PIL import Image
    img = Image.fromarray(img)
    img = img.convert('RGBA')
    L, H = img.size
    color_0 = img.getpixel((2, 2))
    for h in range(H):
        for l in range(L):
            dot = (l, h)
            color_1 = img.getpixel(dot)
            if color_1 == color_0:
                color_1 = color_1[:-1] + (0,)
                img.putpixel(dot, (0, 0, 0, 0))
    return img


def niche_counter(obs,
                  niche_name,
                  cluster_name,
                  niche_label,
                  cluster_label,
                  ) -> list:
    from collections import Counter

    n_cluster = len(cluster_name)
    list_cluster = []
    for i in range(n_cluster):
        list_cluster.append([])

    for index_k in niche_name:
        df_k = obs[obs[niche_label] == index_k]
        dict_k = Counter(df_k[cluster_label].tolist())
        sum = 0
        for val in dict_k.values():
            sum += val
        for key in dict_k.keys():
            dict_k[key] = dict_k[key] / sum
        for k, cluster in enumerate(cluster_name):
            if cluster in dict_k.keys():
                list_cluster[k].append(dict_k[cluster])
            else:
                list_cluster[k].append(0)

    for i in range(n_cluster):
        list_cluster[i] = np.array(list_cluster[i])

    return list_cluster


### function from stlearn

def get_cmap(cmap):
    """Checks inputted cmap string."""
    if cmap == "vega_10_scanpy":
        cmap = palettes.vega_10_scanpy
    elif cmap == "vega_20_scanpy":
        cmap = palettes.vega_20_scanpy
    elif cmap == "default_102":
        cmap = palettes.default_102
    elif cmap == "default_28":
        cmap = palettes.default_28
    elif type(cmap) == str:  # If refers to matplotlib cmap
        cmap_n = plt.get_cmap(cmap).N
        return plt.get_cmap(cmap), cmap_n
    elif type(cmap) == matplotlib.colors.LinearSegmentedColormap:  # already cmap
        cmap_n = cmap.N
        return cmap, cmap_n

    cmap_n = len(cmap)
    cmaps = matplotlib.colors.LinearSegmentedColormap.from_list("", cmap)

    cmap_ = plt.cm.get_cmap(cmaps)

    return cmap_, cmap_n


def check_cmap(cmap):
    """Initialize cmap"""
    scanpy_cmap = ["vega_10_scanpy", "vega_20_scanpy", "default_102", "default_28"]
    stlearn_cmap = ["jana_40", "default"]
    cmap_available = plt.colormaps() + scanpy_cmap + stlearn_cmap
    error_msg = (
            "cmap must be a matplotlib.colors.LinearSegmentedColormap OR"
            "one of these: " + str(cmap_available)
    )
    if type(cmap) == str:
        assert cmap in cmap_available, error_msg
    elif type(cmap) != matplotlib.colors.LinearSegmentedColormap:
        raise Exception(error_msg)

    return cmap


def get_colors(adata, obs_key, cmap=None, label_set=None):
    """
    Retrieves colors if present in adata.uns, if not present then will set
    them as per scanpy & return in order requested.
    """
    # Checking if colors are already set #
    col_key = f"{obs_key}_colors"
    if col_key in adata.uns:
        labels_ordered = adata.obs[obs_key].cat.categories
        colors_ordered = adata.uns[col_key]
    else:  # Colors not already present
        check_cmap(cmap)
        cmap, cmap_n = get_cmap(cmap)

        if not hasattr(adata.obs[obs_key], "cat"):  # Ensure categorical
            adata.obs[obs_key] = adata.obs[obs_key].astype("category")
        labels_ordered = adata.obs[obs_key].cat.categories
        colors_ordered = [
            matplotlib.colors.rgb2hex(cmap(i / (len(labels_ordered) - 1)))
            for i in range(len(labels_ordered))
        ]
        adata.uns[col_key] = colors_ordered

    # Returning the colors of the desired labels in indicated order #
    if label_set is not None:
        colors_ordered = [
            colors_ordered[np.where(labels_ordered == label)[0][0]]
            for label in label_set
        ]

    return colors_ordered


import numpy as np
import pandas as pd
from sklearn import metrics
import scanpy as sc
import ot
from sklearn.decomposition import PCA


def mclust_R(adata, num_cluster, modelNames='EEE', used_obsm='emb_pca', random_seed=2020):
    """\
    Clustering using the mclust algorithm.
    The parameters are the same as those in the R package mclust.
    """
    
    np.random.seed(random_seed)
    import rpy2.robjects as robjects
    robjects.r.library("mclust")

    import rpy2.robjects.numpy2ri
    rpy2.robjects.numpy2ri.activate()
    r_random_seed = robjects.r['set.seed']
    r_random_seed(random_seed)
    rmclust = robjects.r['Mclust']
    
    res = rmclust(rpy2.robjects.numpy2ri.numpy2rpy(adata.obsm[used_obsm]), num_cluster, modelNames)
    mclust_res = np.array(res[-2])

    adata.obs['mclust'] = mclust_res
    adata.obs['mclust'] = adata.obs['mclust'].astype('int')
    adata.obs['mclust'] = adata.obs['mclust'].astype('category')
    return adata

def clustering(adata, n_clusters=7, radius=50, key='emb', method='mclust', start=0.1, end=3.0, increment=0.01, refinement=False):
    """\
    Spatial clustering based the learned representation.

    Parameters
    ----------
    adata : anndata
        AnnData object of scanpy package.
    n_clusters : int, optional
        The number of clusters. The default is 7.
    radius : int, optional
        The number of neighbors considered during refinement. The default is 50.
    key : string, optional
        The key of the learned representation in adata.obsm. The default is 'emb'.
    method : string, optional
        The tool for clustering. Supported tools include 'mclust', 'leiden', and 'louvain'. The default is 'mclust'. 
    start : float
        The start value for searching. The default is 0.1.
    end : float 
        The end value for searching. The default is 3.0.
    increment : float
        The step size to increase. The default is 0.01.   
    refinement : bool, optional
        Refine the predicted labels or not. The default is False.

    Returns
    -------
    None.

    """
    
    pca = PCA(n_components=20, random_state=42) 
    embedding = pca.fit_transform(adata.obsm['emb'].copy())
    adata.obsm['emb_pca'] = embedding
    
    if method == 'mclust':
       adata = mclust_R(adata, used_obsm='emb_pca', num_cluster=n_clusters)
       adata.obs['domain'] = adata.obs['mclust']
    elif method == 'leiden':
       res = search_res(adata, n_clusters, use_rep='emb_pca', method=method, start=start, end=end, increment=increment)
       sc.tl.leiden(adata, random_state=0, resolution=res)
       adata.obs['domain'] = adata.obs['leiden']
    elif method == 'louvain':
       res = search_res(adata, n_clusters, use_rep='emb_pca', method=method, start=start, end=end, increment=increment)
       sc.tl.louvain(adata, random_state=0, resolution=res)
       adata.obs['domain'] = adata.obs['louvain'] 
       
    if refinement:  
       new_type = refine_label(adata, radius, key='domain')
       adata.obs['domain'] = new_type 
       
def refine_label(adata, radius=50, key='label'):
    n_neigh = radius
    new_type = []
    old_type = adata.obs[key].values
    
    #calculate distance
    position = adata.obsm['spatial']
    distance = ot.dist(position, position, metric='euclidean')
           
    n_cell = distance.shape[0]
    
    for i in range(n_cell):
        vec  = distance[i, :]
        index = vec.argsort()
        neigh_type = []
        for j in range(1, n_neigh+1):
            neigh_type.append(old_type[index[j]])
        max_type = max(neigh_type, key=neigh_type.count)
        new_type.append(max_type)
        
    new_type = [str(i) for i in list(new_type)]    
    #adata.obs['label_refined'] = np.array(new_type)
    
    return new_type

def extract_top_value(map_matrix, retain_percent = 0.1): 
    '''\
    Filter out cells with low mapping probability

    Parameters
    ----------
    map_matrix : array
        Mapped matrix with m spots and n cells.
    retain_percent : float, optional
        The percentage of cells to retain. The default is 0.1.

    Returns
    -------
    output : array
        Filtered mapped matrix.

    '''

    #retain top 1% values for each spot
    top_k  = retain_percent * map_matrix.shape[1]
    output = map_matrix * (np.argsort(np.argsort(map_matrix)) >= map_matrix.shape[1] - top_k)
    
    return output 

def construct_cell_type_matrix(adata_sc):
    label = 'cell_type'
    n_type = len(list(adata_sc.obs[label].unique()))
    zeros = np.zeros([adata_sc.n_obs, n_type])
    cell_type = list(adata_sc.obs[label].unique())
    cell_type = [str(s) for s in cell_type]
    cell_type.sort()
    mat = pd.DataFrame(zeros, index=adata_sc.obs_names, columns=cell_type)
    for cell in list(adata_sc.obs_names):
        ctype = adata_sc.obs.loc[cell, label]
        mat.loc[cell, str(ctype)] = 1
    #res = mat.sum()
    return mat

def project_cell_to_spot(adata, adata_sc, retain_percent=0.1):
    '''\
    Project cell types onto ST data using mapped matrix in adata.obsm

    Parameters
    ----------
    adata : anndata
        AnnData object of spatial data.
    adata_sc : anndata
        AnnData object of scRNA-seq reference data.
    retrain_percent: float    
        The percentage of cells to retain. The default is 0.1.
    Returns
    -------
    None.

    '''
    
    # read map matrix 
    map_matrix = adata.obsm['map_matrix']   # spot x cell
   
    # extract top-k values for each spot
    map_matrix = extract_top_value(map_matrix) # filtering by spot
    
    # construct cell type matrix
    matrix_cell_type = construct_cell_type_matrix(adata_sc)
    matrix_cell_type = matrix_cell_type.values
       
    # projection by spot-level
    matrix_projection = map_matrix.dot(matrix_cell_type)
   
    # rename cell types
    cell_type = list(adata_sc.obs['cell_type'].unique())
    cell_type = [str(s) for s in cell_type]
    cell_type.sort()
    #cell_type = [s.replace(' ', '_') for s in cell_type]
    df_projection = pd.DataFrame(matrix_projection, index=adata.obs_names, columns=cell_type)  # spot x cell type
    
    #normalize by row (spot)
    df_projection = df_projection.div(df_projection.sum(axis=1), axis=0).fillna(0)

    #add projection results to adata
    adata.obs[df_projection.columns] = df_projection
    
def search_res(adata, n_clusters, method='leiden', use_rep='emb', start=0.1, end=3.0, increment=0.01):
    '''\
    Searching corresponding resolution according to given cluster number
    
    Parameters
    ----------
    adata : anndata
        AnnData object of spatial data.
    n_clusters : int
        Targetting number of clusters.
    method : string
        Tool for clustering. Supported tools include 'leiden' and 'louvain'. The default is 'leiden'.    
    use_rep : string
        The indicated representation for clustering.
    start : float
        The start value for searching.
    end : float 
        The end value for searching.
    increment : float
        The step size to increase.
        
    Returns
    -------
    res : float
        Resolution.
        
    '''
    print('Searching resolution...')
    label = 0
    sc.pp.neighbors(adata, n_neighbors=50, use_rep=use_rep)
    for res in sorted(list(np.arange(start, end, increment)), reverse=True):
        if method == 'leiden':
           sc.tl.leiden(adata, random_state=0, resolution=res)
           count_unique = len(pd.DataFrame(adata.obs['leiden']).leiden.unique())
           print('resolution={}, cluster number={}'.format(res, count_unique))
        elif method == 'louvain':
           sc.tl.louvain(adata, random_state=0, resolution=res)
           count_unique = len(pd.DataFrame(adata.obs['louvain']).louvain.unique()) 
           print('resolution={}, cluster number={}'.format(res, count_unique))
        if count_unique == n_clusters:
            label = 1
            break

    assert label==1, "Resolution is not found. Please try bigger range or smaller step!." 
       
    return res    



# define color list in SOAPy
from typing import Sequence, Union
from matplotlib.colors import ListedColormap

# default palette
# 10 color
color_list_10 = [
    "#765005", "#0780cf", "#fa6d1d", "#0e2c82", "#b6b51f",
    "#da1f18", "#701866", "#f47a75", "#009db2", "#024b51"
]

# 50 color
color_list_50 = [
    '#5050FFFF', '#CE3D32FF', '#749B58FF', '#F0E685FF', '#466983FF', '#BA6338FF', '#5DB1DDFF',
    '#802268FF', '#6BD76BFF', '#D595A7FF', '#924822FF', '#837B8DFF', '#C75127FF', '#D58F5CFF',
    '#7A65A5FF', '#E4AF69FF', '#3B1B53FF', '#CDDEB7FF', '#612A79FF', '#AE1F63FF', '#E7C76FFF',
    '#5A655EFF', '#CC9900FF', '#99CC00FF', '#A9A9A9FF', '#CC9900FF', '#99CC00FF', '#00D68FFF',
    '#14FFB1FF', '#00CC99FF', '#0099CCFF', '#0A47FFFF', '#4775FFFF', '#FFC20AFF', '#FFD147FF',
    '#990033FF', '#991A00FF', '#996600FF', '#809900FF', '#339900FF', '#00991AFF', '#009966FF',
    '#008099FF', '#003399FF', '#1A0099FF', '#660099FF', '#990080FF', '#D60047FF', '#FF1463FF',
    '#00D68FFF'
]

# default colorbar
cmap_default = 'parula'


def _get_palette(categorical, sort_order: bool = True, palette: Union[Sequence, ListedColormap] = None) -> dict:
    are_all_str = all(map(lambda x: isinstance(x, str), categorical))
    if not are_all_str:
        categorical = str(categorical)

    if sort_order:
        categorical = sorted(categorical)

    if palette is None:
        if len(categorical) <= 10:
            palette = color_list_10
        else:
            palette = color_list_50

    if isinstance(palette, ListedColormap):
        palette = palette.colors

    palette = palette[0: len(categorical)]
    palette = dict(zip(categorical, palette))
    return palette